# Licensed under a 3-clause BSD style license - see LICENSE.rst
import operator

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from astropy import units as u
from astropy.coordinates import Angle
from astropy.tests.helper import assert_quantity_allclose
from astropy.uncertainty import distributions as ds
from astropy.uncertainty.core import Distribution
from astropy.utils import NumpyRNGContext
from astropy.utils.compat.optional_deps import HAS_SCIPY

if HAS_SCIPY:
    from scipy.stats import norm  # pylint: disable=W0611

    SMAD_FACTOR = 1 / norm.ppf(0.75)


class TestInit:
    @classmethod
    def setup_class(cls):
        cls.rates = np.array([1, 5, 30, 400])[:, np.newaxis]
        cls.parr = np.random.poisson(cls.rates, (4, 1000))
        cls.parr_t = np.random.poisson(cls.rates.squeeze(), (1000, 4))

    def test_numpy_init(self):
        # Test that we can initialize directly from a Numpy array
        Distribution(self.parr)

    def test_numpy_init_T(self):
        Distribution(self.parr_t.T)

    def test_quantity_init(self):
        # Test that we can initialize directly from a Quantity
        pq = self.parr << u.ct
        pqd = Distribution(pq)
        assert isinstance(pqd, u.Quantity)
        assert isinstance(pqd, Distribution)
        assert isinstance(pqd.value, Distribution)
        assert_array_equal(pqd.value.distribution, self.parr)

    def test_quantity_init_T(self):
        # Test that we can initialize directly from a Quantity
        pq = self.parr_t << u.ct
        Distribution(pq.T)

    def test_quantity_init_with_distribution(self):
        # Test that we can initialize a Quantity from a Distribution.
        pd = Distribution(self.parr)
        qpd = pd << u.ct
        assert isinstance(qpd, u.Quantity)
        assert isinstance(qpd, Distribution)
        assert qpd.unit == u.ct
        assert_array_equal(qpd.value.distribution, pd.distribution.astype(float))


def test_init_scalar():
    parr = np.random.poisson(np.array([1, 5, 30, 400])[:, np.newaxis], (4, 1000))
    with pytest.raises(
        TypeError, match=r"Attempted to initialize a Distribution with a scalar"
    ):
        Distribution(parr.ravel()[0])


class TestDistributionStatistics:
    def setup_class(self):
        with NumpyRNGContext(12345):
            self.data = np.random.normal(
                np.array([1, 2, 3, 4])[:, np.newaxis],
                np.array([3, 2, 4, 5])[:, np.newaxis],
                (4, 10000),
            )

        self.distr = Distribution(self.data * u.kpc)

    def test_shape(self):
        # Distribution shape
        assert self.distr.shape == (4,)
        assert self.distr.distribution.shape == (4, 10000)

    def test_size(self):
        # Total number of values
        assert self.distr.size == 4
        assert self.distr.distribution.size == 40000

    def test_n_samples(self):
        # Number of samples
        assert self.distr.n_samples == 10000

    def test_n_distr(self):
        assert self.distr.shape == (4,)

    def test_pdf_mean(self):
        # Mean of each PDF
        expected = np.mean(self.data, axis=-1) * self.distr.unit
        pdf_mean = self.distr.pdf_mean()
        assert_quantity_allclose(pdf_mean, expected)
        assert_quantity_allclose(pdf_mean, [1, 2, 3, 4] * self.distr.unit, rtol=0.05)

        # make sure the right type comes out - should be a Quantity because it's
        # now a summary statistic
        assert not isinstance(pdf_mean, Distribution)
        assert isinstance(pdf_mean, u.Quantity)

        # Check with out argument.
        out = pdf_mean * 0.0
        pdf_mean2 = self.distr.pdf_mean(out=out)
        assert pdf_mean2 is out
        assert np.all(pdf_mean2 == pdf_mean)

    def test_pdf_std(self):
        # Standard deviation of each PDF
        expected = np.std(self.data, axis=-1) * self.distr.unit
        pdf_std = self.distr.pdf_std()
        assert_quantity_allclose(pdf_std, expected)
        assert_quantity_allclose(pdf_std, [3, 2, 4, 5] * self.distr.unit, rtol=0.05)

        # make sure the right type comes out - should be a Quantity because it's
        # now a summary statistic
        assert not isinstance(pdf_std, Distribution)
        assert isinstance(pdf_std, u.Quantity)

        # Check with proper ddof, using out argument.
        out = pdf_std * 0.0
        expected = np.std(self.data, axis=-1, ddof=1) * self.distr.unit
        pdf_std2 = self.distr.pdf_std(ddof=1, out=out)
        assert pdf_std2 is out
        assert np.all(pdf_std2 == expected)

    def test_pdf_var(self):
        # Variance of each PDF
        expected = np.var(self.data, axis=-1) * self.distr.unit**2
        pdf_var = self.distr.pdf_var()
        assert_quantity_allclose(pdf_var, expected)
        assert_quantity_allclose(pdf_var, [9, 4, 16, 25] * self.distr.unit**2, rtol=0.1)

        # make sure the right type comes out - should be a Quantity because it's
        # now a summary statistic
        assert not isinstance(pdf_var, Distribution)
        assert isinstance(pdf_var, u.Quantity)

        # Check with proper ddof, using out argument.
        out = pdf_var * 0.0
        expected = np.var(self.data, axis=-1, ddof=1) * self.distr.unit**2
        pdf_var2 = self.distr.pdf_var(ddof=1, out=out)
        assert pdf_var2 is out
        assert np.all(pdf_var2 == expected)

    def test_pdf_median(self):
        # Median of each PDF
        expected = np.median(self.data, axis=-1) * self.distr.unit
        pdf_median = self.distr.pdf_median()
        assert_quantity_allclose(pdf_median, expected)
        assert_quantity_allclose(pdf_median, [1, 2, 3, 4] * self.distr.unit, rtol=0.1)

        # make sure the right type comes out - should be a Quantity because it's
        # now a summary statistic
        assert not isinstance(pdf_median, Distribution)
        assert isinstance(pdf_median, u.Quantity)

        # Check with out argument.
        out = pdf_median * 0.0
        pdf_median2 = self.distr.pdf_median(out=out)
        assert pdf_median2 is out
        assert np.all(pdf_median2 == expected)

    @pytest.mark.skipif(not HAS_SCIPY, reason="no scipy")
    def test_pdf_mad_smad(self):
        # Median absolute deviation of each PDF
        median = np.median(self.data, axis=-1, keepdims=True)
        expected = np.median(np.abs(self.data - median), axis=-1) * self.distr.unit
        pdf_mad = self.distr.pdf_mad()
        assert_quantity_allclose(pdf_mad, expected)
        pdf_smad = self.distr.pdf_smad()
        assert_quantity_allclose(pdf_smad, pdf_mad * SMAD_FACTOR, rtol=1e-5)
        assert_quantity_allclose(pdf_smad, [3, 2, 4, 5] * self.distr.unit, rtol=0.05)

        # make sure the right type comes out - should be a Quantity because it's
        # now a summary statistic
        assert not isinstance(pdf_mad, Distribution)
        assert isinstance(pdf_mad, u.Quantity)
        assert not isinstance(pdf_smad, Distribution)
        assert isinstance(pdf_smad, u.Quantity)

        # Check out argument for smad (which checks mad too).
        out = pdf_smad * 0.0
        pdf_smad2 = self.distr.pdf_smad(out=out)
        assert pdf_smad2 is out
        assert np.all(pdf_smad2 == pdf_smad)

    def test_percentile(self):
        expected = np.percentile(self.data, [10, 50, 90], axis=-1) * self.distr.unit
        percs = self.distr.pdf_percentiles([10, 50, 90])
        assert_quantity_allclose(percs, expected)
        assert percs.shape == (3, 4)

        # make sure the right type comes out - should be a Quantity because it's
        # now a summary statistic
        assert not isinstance(percs, Distribution)
        assert isinstance(percs, u.Quantity)

    def test_add_quantity(self):
        distrplus = self.distr + [2000, 0, 0, 500] * u.pc
        expected = (
            np.median(self.data, axis=-1) + np.array([2, 0, 0, 0.5])
        ) * self.distr.unit
        assert_quantity_allclose(distrplus.pdf_median(), expected)
        expected = np.var(self.data, axis=-1) * self.distr.unit**2
        assert_quantity_allclose(distrplus.pdf_var(), expected)

    def test_add_distribution(self):
        another_data = (
            np.random.randn(4, 10000) * np.array([1000, 0.01, 80, 10])[:, np.newaxis]
            + np.array([2000, 0, 0, 500])[:, np.newaxis]
        )
        # another_data is in pc, but main distr is in kpc
        another_distr = Distribution(another_data * u.pc)
        combined_distr = self.distr + another_distr

        expected = np.median(self.data + another_data / 1000, axis=-1) * self.distr.unit
        assert_quantity_allclose(combined_distr.pdf_median(), expected)

        expected = np.var(self.data + another_data / 1000, axis=-1) * self.distr.unit**2
        assert_quantity_allclose(combined_distr.pdf_var(), expected)


def test_helper_normal_samples():
    centerq = [1, 5, 30, 400] * u.kpc

    with NumpyRNGContext(12345):
        n_dist = ds.normal(centerq, std=[0.2, 1.5, 4, 1] * u.kpc, n_samples=100)
        assert n_dist.distribution.shape == (4, 100)
        assert n_dist.shape == (4,)
        assert n_dist.unit == u.kpc
        assert np.all(n_dist.pdf_std() > 100 * u.pc)

        n_dist2 = ds.normal(centerq, std=[0.2, 1.5, 4, 1] * u.pc, n_samples=20000)
        assert n_dist2.distribution.shape == (4, 20000)
        assert n_dist2.shape == (4,)
        assert n_dist2.unit == u.kpc
        assert np.all(n_dist2.pdf_std() < 100 * u.pc)


def test_helper_poisson_samples():
    centerqcounts = [1, 5, 30, 400] * u.count

    with NumpyRNGContext(12345):
        p_dist = ds.poisson(centerqcounts, n_samples=100)
        assert p_dist.shape == (4,)
        assert p_dist.distribution.shape == (4, 100)
        assert p_dist.unit == u.count
        p_min = np.min(p_dist)
        assert isinstance(p_min, Distribution)
        assert p_min.shape == ()
        assert np.all(p_min >= 0)
        assert np.all(np.abs(p_dist.pdf_mean() - centerqcounts) < centerqcounts)


def test_helper_uniform_samples():
    udist = ds.uniform(lower=[1, 2] * u.kpc, upper=[3, 4] * u.kpc, n_samples=1000)
    assert udist.shape == (2,)
    assert udist.distribution.shape == (2, 1000)
    assert np.all(np.min(udist.distribution, axis=-1) > [1, 2] * u.kpc)
    assert np.all(np.max(udist.distribution, axis=-1) < [3, 4] * u.kpc)

    # try the alternative creator
    udist = ds.uniform(center=[1, 3, 2] * u.pc, width=[5, 4, 3] * u.pc, n_samples=1000)
    assert udist.shape == (3,)
    assert udist.distribution.shape == (3, 1000)
    assert np.all(np.min(udist.distribution, axis=-1) > [-1.5, 1, 0.5] * u.pc)
    assert np.all(np.max(udist.distribution, axis=-1) < [3.5, 5, 3.5] * u.pc)


def test_helper_normal_exact():
    pytest.skip("distribution stretch goal not yet implemented")
    centerq = [1, 5, 30, 400] * u.kpc
    ds.normal(centerq, std=[0.2, 1.5, 4, 1] * u.kpc)
    ds.normal(centerq, var=[0.04, 2.25, 16, 1] * u.kpc**2)
    ds.normal(centerq, ivar=[25, 0.44444444, 0.625, 1] * u.kpc**-2)


def test_helper_poisson_exact():
    pytest.skip("distribution stretch goal not yet implemented")
    centerq = [1, 5, 30, 400] * u.one
    ds.poisson(centerq, n_samples=1000)

    with pytest.raises(
        u.UnitsError,
        match=r"Poisson distribution can only be computed for dimensionless quantities",
    ):
        centerq = [1, 5, 30, 400] * u.kpc
        ds.poisson(centerq, n_samples=1000)


def test_reprs():
    darr = np.arange(30).reshape(3, 10)
    distr = Distribution(darr * u.kpc)

    assert "n_samples=10" in repr(distr)
    assert "n_samples=10" in str(distr)

    assert r"n_{\rm samp}=10" in distr._repr_latex_()


@pytest.mark.parametrize(
    "func, kws",
    [
        (ds.normal, {"center": 0, "std": 2}),
        (ds.uniform, {"lower": 0, "upper": 2}),
        (ds.poisson, {"center": 2}),
        (ds.normal, {"center": 0 * u.count, "std": 2 * u.count}),
        (ds.uniform, {"lower": 0 * u.count, "upper": 2 * u.count}),
        (ds.poisson, {"center": 2 * u.count}),
    ],
)
def test_wrong_kw_fails(func, kws):
    with pytest.raises(TypeError, match="missing 1 required"):
        kw_temp = kws.copy()
        kw_temp["n_sample"] = 100  # note the missing "s"
        assert func(**kw_temp).n_samples == 100
    kw_temp = kws.copy()
    kw_temp["n_samples"] = 100
    assert func(**kw_temp).n_samples == 100


def test_index_assignment_quantity():
    arr = np.random.randn(2, 1000)
    distr = Distribution(arr * u.kpc)
    d1q, d2q = distr
    assert isinstance(d1q, Distribution)
    assert isinstance(d2q, Distribution)

    ndistr = ds.normal(center=[1, 2] * u.kpc, std=[3, 4] * u.kpc, n_samples=1000)
    n1, n2 = ndistr
    assert isinstance(n1, ds.Distribution)
    assert isinstance(n2, ds.Distribution)


def test_index_assignment_array():
    arr = np.random.randn(2, 1000)
    distr = Distribution(arr)
    d1a, d2a = distr
    assert isinstance(d1a, Distribution)
    assert isinstance(d2a, Distribution)

    ndistr = ds.normal(center=[1, 2], std=[3, 4], n_samples=1000)
    n1, n2 = ndistr
    assert isinstance(n1, ds.Distribution)
    assert isinstance(n2, ds.Distribution)


def test_histogram():
    arr = np.random.randn(2, 3, 1000)
    distr = Distribution(arr)

    hist, bins = distr.pdf_histogram(bins=10)
    assert hist.shape == (2, 3, 10)
    assert bins.shape == (2, 3, 11)


def test_array_repr_latex():
    # as of this writing ndarray does not have a _repr_latex_, and this test
    # ensure distributions account for that. However, if in the future ndarray
    # gets a _repr_latex_, we can skip this.

    arr = np.random.randn(4, 1000)

    if hasattr(arr, "_repr_latex_"):
        pytest.skip("in this version of numpy, ndarray has a _repr_latex_")

    distr = Distribution(arr)
    assert distr._repr_latex_() is None


def test_distr_to():
    distr = ds.normal(10 * u.cm, n_samples=100, std=1 * u.cm)
    todistr = distr.to(u.m)
    assert_quantity_allclose(distr.pdf_mean().to(u.m), todistr.pdf_mean())


def test_distr_noq_to():
    # this is an array distribution not a quantity
    distr = ds.normal(10, n_samples=100, std=1)
    with pytest.raises(AttributeError):
        distr.to(u.m)


def test_distr_to_value():
    distr = ds.normal(10 * u.cm, n_samples=100, std=1 * u.cm)
    tovdistr = distr.to_value(u.m)
    assert np.allclose(distr.pdf_mean().to_value(u.m), tovdistr.pdf_mean())


def test_distr_noq_to_value():
    distr = ds.normal(10, n_samples=100, std=1)
    with pytest.raises(AttributeError):
        distr.to_value(u.m)


def test_distr_angle():
    # Check that Quantity subclasses decay to Quantity appropriately.
    distr = Distribution([2.0, 3.0, 4.0])
    ad = Angle(distr, "deg")
    ad_plus_ad = ad + ad
    assert isinstance(ad_plus_ad, Angle)
    assert isinstance(ad_plus_ad, Distribution)

    ad_times_ad = ad * ad
    assert not isinstance(ad_times_ad, Angle)
    assert isinstance(ad_times_ad, u.Quantity)
    assert isinstance(ad_times_ad, Distribution)

    ad += ad
    assert isinstance(ad, Angle)
    assert isinstance(ad, Distribution)
    assert_array_equal(ad.distribution, ad_plus_ad.distribution)

    with pytest.raises(u.UnitTypeError):
        ad *= ad


def test_distr_angle_view_as_quantity():
    # Check that Quantity subclasses decay to Quantity appropriately.
    distr = Distribution([2.0, 3.0, 4.0])
    ad = Angle(distr, "deg")
    qd = ad.view(u.Quantity)
    assert not isinstance(qd, Angle)
    assert isinstance(qd, u.Quantity)
    assert isinstance(qd, Distribution)
    # View directly as DistributionQuantity class.
    qd2 = ad.view(qd.__class__)
    assert not isinstance(qd2, Angle)
    assert isinstance(qd2, u.Quantity)
    assert isinstance(qd2, Distribution)
    assert_array_equal(qd2.distribution, qd.distribution)
    qd3 = ad.view(qd.dtype, qd.__class__)
    assert not isinstance(qd3, Angle)
    assert isinstance(qd3, u.Quantity)
    assert isinstance(qd3, Distribution)
    assert_array_equal(qd3.distribution, qd.distribution)
    # Simple view with no arguments
    qd4 = qd3.view()
    assert qd4.__class__ is qd3.__class__
    assert np.may_share_memory(qd4, qd3)


def test_distr_view_different_dtype1():
    # Viewing with a new dtype should follow the same rules as for a
    # regular array with the same dtype.
    c = Distribution([2.0j, 3.0, 4.0j])
    r = c.view("2f8")
    assert r.shape == c.shape + (2,)
    assert np.may_share_memory(r, c)
    expected = np.moveaxis(c.distribution.view("2f8"), -2, -1)
    assert_array_equal(r.distribution, expected)
    c2 = r.view("c16")
    assert_array_equal(c2.distribution, c.distribution)
    assert np.may_share_memory(c2, c)


def test_distr_view_different_dtype2():
    # Viewing with a new dtype should follow the same rules as for a
    # regular array with the same dtype.
    uint32 = Distribution(
        np.array([[0x01020304, 0x05060708], [0x11121314, 0x15161718]], dtype="u4")
    )
    uint8 = uint32.view("4u1")
    assert uint8.shape == uint32.shape + (4,)
    assert np.may_share_memory(uint8, uint32)
    expected = np.moveaxis(uint32.distribution.view("4u1"), -2, -1)
    assert_array_equal(uint8.distribution, expected)
    uint32_2 = uint8.view("u4")
    assert np.may_share_memory(uint32_2, uint32)
    assert_array_equal(uint32_2.distribution, uint32.distribution)
    uint8_2 = uint8.T
    with pytest.raises(ValueError, match="last axis must be contiguous"):
        uint8_2.view("u4")


def test_distr_cannot_view_new_dtype():
    # A Distribution has a very specific structured dtype with just one
    # element that holds the array of samples.  As it is not clear what
    # to do with a view as a new dtype, we just error on it.
    # TODO: with a lot of thought, this restriction can likely be relaxed.
    distr = Distribution([2.0, 3.0, 4.0])
    with pytest.raises(ValueError, match="can only be viewed"):
        distr.view(np.dtype("2i8"))

    with pytest.raises(ValueError, match="can only be viewed"):
        distr.view(np.dtype("2i8"), distr.__class__)

    # Check subclass just in case.
    ad = Angle(distr, "deg")
    with pytest.raises(ValueError, match="can only be viewed"):
        ad.view(np.dtype("2i8"))
    with pytest.raises(ValueError, match="can only be viewed"):
        ad.view("2i8", distr.__class__)


def test_scalar_quantity_distribution():
    # Regression test for gh-12336
    angles = Distribution([90.0, 30.0, 0.0] * u.deg)
    sin_angles = np.sin(angles)  # This failed in 4.3.
    assert isinstance(sin_angles, Distribution)
    assert isinstance(sin_angles, u.Quantity)
    assert_array_equal(sin_angles, Distribution(np.sin([90.0, 30.0, 0.0] * u.deg)))


@pytest.mark.parametrize("op", [operator.eq, operator.ne, operator.gt])
class TestComparison:
    @classmethod
    def setup_class(cls):
        cls.d = Distribution([90.0, 30.0, 0.0])

        class Override:
            __array_ufunc__ = None

            def __eq__(self, other):
                return "eq"

            def __ne__(self, other):
                return "ne"

            def __lt__(self, other):
                return "gt"  # Since it is called for the reverse of gt

        cls.override = Override()

    def test_distribution_can_be_compared_to_non_distribution(self, op):
        result = op(self.d, 0.0)
        assert_array_equal(result, Distribution(op(self.d.distribution, 0.0)))

    def test_distribution_comparison_defers_correctly(self, op):
        result = op(self.d, self.override)
        assert result == op.__name__


class TestSetItemWithSelection:
    def test_setitem(self):
        d = Distribution([90.0, 30.0, 0.0])
        d[d > 50] = 0.0
        assert_array_equal(d, Distribution([0.0, 30.0, 0.0]))

    def test_inplace_operation(self):
        d = Distribution([90.0, 30.0, 0.0])
        d[d > 50] *= -1.0
        assert_array_equal(d, Distribution([-90.0, 30.0, 0.0]))


ADVANCED_INDICES = [
    (np.array([[1, 2], [0, 1]]), np.array([[1, 3], [0, 2]])),
    # Same advanced index, but also using negative numbers
    (np.array([[-2, -1], [0, -2]]), np.array([[1, -1], [-4, -2]])),
]


class TestGetSetItemAdvancedIndex:
    @classmethod
    def setup_class(cls):
        cls.distribution = np.arange(60.0).reshape(3, 4, 5)
        cls.d = Distribution(cls.distribution)

    def test_setup(self):
        ai1, ai2 = ADVANCED_INDICES[:2]
        # Check that the first two indices produce the same output.
        assert_array_equal(self.distribution[ai1], self.distribution[ai2])

    @pytest.mark.parametrize("item", ADVANCED_INDICES)
    def test_getitem(self, item):
        v = self.d[item]
        assert v.shape == item[0].shape
        assert_array_equal(v.distribution, self.distribution[item])

    @pytest.mark.parametrize("item", [([0, 4],), ([0], [0], [0])])
    def test_getitem_bad(self, item):
        with pytest.raises(IndexError):
            self.d[item]

    @pytest.mark.parametrize("item", ADVANCED_INDICES)
    def test_setitem(self, item):
        d = self.d.copy()
        d[item] = 0.0
        distribution = self.distribution.copy()
        distribution[item] = 0.0
        assert_array_equal(d.distribution, distribution)
        d[item] = self.d[item]
        assert_array_equal(d.distribution, self.distribution)


class TestQuantityDistributionGetSetItemAdvancedIndex(TestGetSetItemAdvancedIndex):
    @classmethod
    def setup_class(cls):
        cls.distribution = np.arange(60.0).reshape(3, 4, 5) << u.m
        cls.d = Distribution(cls.distribution)


class StructuredDtypeBase:
    @classmethod
    def setup_class(cls):
        cls.dtype = np.dtype([("a", "f8"), ("b", "(2,2)f8")])
        data = np.arange(5.0) + (np.arange(60.0) * 10).reshape(3, 4, 5, 1)
        cls.distribution = data.view(cls.dtype).reshape(3, 4, 5)
        cls.d = Distribution(cls.distribution)


class TestStructuredQuantityDistributionInit(StructuredDtypeBase):
    @classmethod
    def setup_class(cls):
        super().setup_class()
        cls.unit = u.Unit("km, m")
        cls.d_unit = cls.unit

    def test_init_via_structured_samples(self):
        distribution = self.distribution << self.unit
        d = Distribution(distribution)
        assert d.unit == self.d_unit
        assert_array_equal(d.distribution, distribution)
        assert_array_equal(d.value.distribution, self.distribution)

    def test_init_via_structured_distribution(self):
        d = self.d << self.unit
        assert d.unit == self.d_unit


class TestStructuredAdvancedIndex(StructuredDtypeBase, TestGetSetItemAdvancedIndex):
    def test_init(self):
        assert self.d.shape == (3, 4)
        assert self.d.n_samples == 5
        assert_array_equal(self.d.distribution, self.distribution)


class TestStructuredDistribution(StructuredDtypeBase):
    @classmethod
    def setup_class(cls):
        super().setup_class()
        cls.item = (0.0, [[-1.0, -2.0], [-3.0, -4.0]])
        cls.b_item = [[-1.0, -2.0], [-3.0, -4.0]]

    @pytest.mark.parametrize("item", [-2, slice(1, 3), "a", "b"])
    def test_getitem(self, item):
        d_i = self.d[item]
        assert isinstance(d_i, Distribution)
        if item in self.dtype.names:
            assert d_i.shape == self.d.shape + self.dtype[item].shape
        else:
            assert d_i.shape == np.ones(self.d.shape)[item].shape
        assert np.may_share_memory(d_i, self.d)
        expected = self.distribution[item]
        if isinstance(item, str):
            # Sample axis should always be at the end.
            expected = np.moveaxis(self.distribution[item], self.d.ndim, -1)
        assert_array_equal(d_i.distribution, expected)

    @pytest.mark.parametrize("item", [1, slice(0, 2)])
    def test_setitem_index_slice(self, item):
        d = self.d.copy()
        distribution = self.distribution.copy()
        value = self.item
        d[item] = value
        distribution[item] = value
        assert_array_equal(d.distribution, distribution)
        d[item] = self.d[item]
        assert_array_equal(d.distribution, self.distribution)

    @pytest.mark.parametrize("item", ["a", "b"])
    def test_setitem_field(self, item):
        d = self.d.copy()
        d[item] = 0.0
        assert_array_equal(d.distribution[item], np.zeros_like(self.distribution[item]))
        if item == "b":
            value = self.b_item  # selected to be a bit tricky.
            d[item] = value
            assert_array_equal(
                d.distribution[item], np.full_like(self.distribution[item], value)
            )
        d[item] = self.d[item] * 2.0
        assert_array_equal(d.distribution[item], self.distribution[item] * 2)


class TestStructuredQuantityDistribution(TestStructuredDistribution):
    @classmethod
    def setup_class(cls):
        super().setup_class()
        cls.distribution = cls.distribution * u.Unit("km,m")
        cls.d = cls.d * u.Unit("km,m")
        cls.item = cls.item * u.Unit("Mm,km")
        cls.b_item = cls.b_item * u.km



# Licensed under a 3-clause BSD style license - see LICENSE.rst


from io import StringIO

import numpy as np
import pytest

from astropy.io import ascii


def test_types_from_dat():
    converters = {"a": [ascii.convert_numpy(float)], "e": [ascii.convert_numpy(str)]}

    dat = ascii.read(
        ["a b c d e", "1 1 cat 2.1 4.2"], format="basic", converters=converters
    )

    assert dat["a"].dtype.kind == "f"
    assert dat["b"].dtype.kind == "i"
    assert dat["c"].dtype.kind in ("S", "U")
    assert dat["d"].dtype.kind == "f"
    assert dat["e"].dtype.kind in ("S", "U")


def test_rdb_write_types():
    dat = ascii.read(["a b c d", "1 1.0 cat 2.1"], format="basic")
    out = StringIO()
    ascii.write(dat, out, format="rdb")
    outs = out.getvalue().splitlines()
    assert outs[1] == "N\tN\tS\tN"


def test_ipac_read_types():
    table = r"""\
|     ra   |    dec   |   sai   |-----v2---|    sptype        |
|    real  |   float  |   l     |    real  |     char         |
|    unit  |   unit   |   unit  |    unit  |     ergs         |
|    null  |   null   |   null  |    null  |     -999         |
   2.09708   2956        73765    2.06000   B8IVpMnHg
"""
    reader = ascii.get_reader(reader_cls=ascii.Ipac)
    reader.read(table)
    types = [
        ascii.FloatType,
        ascii.FloatType,
        ascii.IntType,
        ascii.FloatType,
        ascii.StrType,
    ]
    for col, expected_type in zip(reader.cols, types):
        assert col.type == expected_type


def test_ipac_read_long_columns():
    """Test for https://github.com/astropy/astropy/issues/15989"""
    test_data = """\
|              oid|expid|cadence|
|             long|    l|    int|
 90000000000000001   123       1
"""
    dat = ascii.read(test_data, format="ipac")

    # assert oid, as a long column, is int64
    oid = dat["oid"]
    assert oid[0] == 90000000000000001
    assert oid.dtype.kind == "i"
    assert oid.dtype.itemsize == 8

    # expid is declared as a long column,
    # the type needs to be int64, even though all
    # the values are within int32 range
    expid = dat["expid"]
    assert expid[0] == 123
    assert expid.dtype.kind == "i"
    assert expid.dtype.itemsize == 8


@pytest.mark.parametrize("fast", [False, "force"])
def test_large_integers(fast):
    # case https://github.com/astropy/astropy/issues/5744
    dat = ascii.read(
        """\
num,cadence
110000000000000001,1
12345,2
""",
        fast_reader=fast,
    )
    # assert `num`` is exactly 64bit int
    assert dat["num"].dtype.kind == "i"
    assert dat["num"].dtype.itemsize == 8
    assert dat["num"][0] == 110000000000000001

    # assert `cadence` is exactly 64bit int, even though the values are small numbers.
    assert dat["cadence"].dtype.kind == "i"
    assert dat["cadence"].dtype.itemsize == 8


def test_col_dtype_in_custom_class():
    """Test code in BaseOutputter._convert_vals to handle Column.dtype
    attribute. See discussion in #11895."""
    dtypes = [np.float32, np.int8, np.int16]

    class TestDtypeHeader(ascii.BasicHeader):
        def get_cols(self, lines):
            super().get_cols(lines)
            for col, dtype in zip(self.cols, dtypes):
                col.dtype = dtype

    class TestDtype(ascii.Basic):
        """
        Basic table Data reader_cls with data type alternating float32, int8
        """

        header_class = TestDtypeHeader

    txt = """
    a b c
    1 2 3
    """
    reader = ascii.get_reader(TestDtype)
    t = reader.read(txt)
    for col, dtype in zip(t.itercols(), dtypes):
        assert col.dtype.type is dtype



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Astropy cosmology flrw module.

.. deprecated:: 7.1

    This module is deprecated and will be removed in a future version. All the public
    classes and functions have been and will continue to be available in the
    :mod:`~astropy.cosmology` module.

"""

__all__ = [  # noqa: RUF100, RUF022, F822
    # base
    "FLRW",
    "FlatFLRWMixin",
    # lambdacdm
    "LambdaCDM",
    "FlatLambdaCDM",
    # w0cdm
    "wCDM",
    "FlatwCDM",
    # w0wacdm
    "w0waCDM",
    "Flatw0waCDM",
    # w0wzcdm
    "w0wzCDM",
    "Flatw0wzCDM",
    # wpwazpcdm
    "wpwaCDM",
    "FlatwpwaCDM",
]

import sys
import warnings
from typing import Any


def __getattr__(name: str) -> Any:
    if name not in __all__:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}.")

    from ._src import flrw

    obj = getattr(flrw, name)

    setattr(sys.modules[__name__], name, obj)

    warnings.warn(
        "The module `astropy.cosmology.flrw` is deprecated since v7.1 and will be "
        "removed in a future version. Import from `astropy.cosmology` instead.",
        category=DeprecationWarning,
    )

    return obj



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Can `astropy.io.fits.open` access (remote) data using the fsspec package?"""

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_array_equal

from astropy.io import fits
from astropy.nddata import Cutout2D
from astropy.table import Table
from astropy.utils.compat.optional_deps import HAS_FSSPEC, HAS_S3FS
from astropy.utils.data import get_pkg_data_filename

if HAS_FSSPEC:
    import fsspec


@pytest.mark.skipif(not HAS_FSSPEC, reason="requires fsspec")
def test_fsspec_local():
    """Can we use fsspec to read a local file?"""
    fn = get_pkg_data_filename("data/test0.fits")
    with fits.open(fn) as hdulist_classic:
        with fits.open(fn, use_fsspec=True) as hdulist_fsspec:
            assert_array_equal(hdulist_classic[2].data, hdulist_fsspec[2].data)
            assert_array_equal(
                hdulist_classic[2].section[3:5], hdulist_fsspec[2].section[3:5]
            )


@pytest.mark.skipif(not HAS_FSSPEC, reason="requires fsspec")
def test_fsspec_local_write(tmp_path):
    """Can we write to a local file that was opened using fsspec?"""
    fn = get_pkg_data_filename("data/test0.fits")
    fn_tmp = tmp_path / "tmp.fits"
    with fits.open(fn, use_fsspec=True) as hdul:
        # writing to a section is never allowed
        with pytest.raises(TypeError):
            hdul[1].section[0, 0] = -999
        # however writing to .data should work
        hdul[1].data[2, 3] = -999
        assert hdul[1].data[2, 3] == -999
        hdul.writeto(fn_tmp)

    # Is the new value present when we re-open the file?
    with fits.open(fn_tmp) as hdul:
        assert hdul[1].data[2, 3] == -999


@pytest.mark.skipif(not HAS_FSSPEC, reason="requires fsspec")
def test_fsspec_cutout2d():
    """Does Cutout2D work with data loaded lazily using fsspec and .section?"""
    fn = get_pkg_data_filename("data/test0.fits")
    with fits.open(fn, use_fsspec=True) as hdul:
        position = (10, 20)
        size = (2, 3)
        cutout1 = Cutout2D(hdul[1].data, position, size)
        cutout2 = Cutout2D(hdul[1].section, position, size)
        assert_allclose(cutout1.data, cutout2.data)


@pytest.mark.skipif(not HAS_FSSPEC, reason="requires fsspec")
def test_fsspec_compressed():
    """Does fsspec support compressed data correctly?"""
    # comp.fits[1] is a compressed image with shape (440, 300)
    fn = get_pkg_data_filename("data/comp.fits")
    with fits.open(fn, use_fsspec=True) as hdul:
        # The .data attribute should work as normal
        assert hdul[1].data[0, 0] == 7
        # And the .section attribute should work too
        assert hdul[1].section[0, 0] == 7


@pytest.mark.remote_data
class TestFsspecRemote:
    """Test obtaining cutouts from FITS files via HTTP (from MAST) and S3 (from Amazon)."""

    def setup_class(self):
        # The test file (ibxl50020_jif.fits) is a Hubble jitter FITS file (*.jif)
        # rather than a real image, because jitter files are less likely to
        # change due to reprocessing.
        self.http_url = "https://mast.stsci.edu/api/v0.1/Download/file/?uri=mast:HST/product/ibxl50020_jif.fits"
        self.s3_uri = "s3://stpubdata/hst/public/ibxl/ibxl50020/ibxl50020_jif.fits"
        self.s3_uri_table = "s3://stpubdata/jwst/public/jw02565/jw02565001001/jw02565001001_03101_00002_nrs2_x1d.fits"
        # Random slice was selected for testing:
        self.slice = (slice(31, 33), slice(27, 30))
        # The expected cutout array below was obtained by downloading the URIs
        # listed above to a local path and and executing:
        # with fits.open(local_path) as hdul:
        #     expected_cutout = hdul[1].data[31:33, 27:30]
        self.expected_cutout = np.array([[24, 88, 228], [35, 132, 305]], dtype=np.int32)

    @pytest.mark.skipif(not HAS_FSSPEC, reason="requires fsspec")
    def test_fsspec_http(self):
        """Can we use fsspec to open a remote FITS file via http?"""
        with fits.open(self.http_url, use_fsspec=True) as hdul:
            # Do we retrieve the expected array?
            assert_array_equal(hdul[1].section[self.slice], self.expected_cutout)
            # The file has multiple extensions which are not yet downloaded;
            # the repr and string representation should reflect this.
            assert "partially read" in repr(hdul)
            assert "partially read" in str(hdul)

        # Can the user also pass an fsspec file object directly to fits open?
        with fsspec.open(self.http_url) as fileobj:
            with fits.open(fileobj) as hdul2:
                assert_array_equal(hdul2[1].section[self.slice], self.expected_cutout)
                assert "partially read" in repr(hdul)
                assert "partially read" in str(hdul)

    @pytest.mark.skipif(not HAS_S3FS, reason="requires s3fs")
    def test_fsspec_s3(self):
        """Can we use fsspec to open a FITS file in a public Amazon S3 bucket?"""
        with fits.open(
            self.s3_uri, fsspec_kwargs={"anon": True}
        ) as hdul:  # s3:// paths should default to use_fsspec=True
            # Do we retrieve the expected array?
            assert_array_equal(hdul[1].section[self.slice], self.expected_cutout)
            # The file has multiple extensions which are not yet downloaded;
            # the repr and string representation should reflect this.
            assert "partially read" in repr(hdul)
            assert "partially read" in str(hdul)

        # Can the user also pass an fsspec file object directly to fits open?
        with fsspec.open(self.s3_uri, anon=True) as fileobj:
            with fits.open(fileobj) as hdul2:
                assert_array_equal(hdul2[1].section[self.slice], self.expected_cutout)
                assert "partially read" in repr(hdul)
                assert "partially read" in str(hdul)

    @pytest.mark.skipif(not HAS_S3FS, reason="requires s3fs")
    def test_fsspec_s3_table(self):
        """Can we use fsspec to open a FITS table in a public Amazon S3 bucket?"""

        # s3:// paths should default to use_fsspec=True
        with fits.open(self.s3_uri_table, fsspec_kwargs={"anon": True}) as hdul:
            t1 = Table(hdul[1].data)

        assert len(t1) == 424
        assert len(t1.colnames) == 18

        t2 = Table.read(
            self.s3_uri_table, fsspec_kwargs={"anon": True}, format="fits", hdu=1
        )

        assert len(t2) == 424
        assert t1.colnames == t2.colnames



from packaging.version import Version

try:
    from ._version import version
except ImportError:
    import warnings

    warnings.warn(
        f"could not determine {__name__.split('.')[0]} package version; "
        "this indicates a broken installation"
    )
    del warnings

    version = "0.0.0"


# We use Version to define major, minor, micro, but ignore any suffixes.
def split_version(version):
    pieces = [0, 0, 0]

    try:
        v = Version(version)
        pieces = [v.major, v.minor, v.micro]

    except Exception:
        pass

    return pieces


major, minor, bugfix = split_version(version)

del split_version  # clean up namespace.

release = "dev" not in version



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This subpackage contains statistical tools provided for or used by Astropy.

While the `scipy.stats` package contains a wide range of statistical
tools, it is a general-purpose package, and is missing some that are
particularly useful to astronomy or are used in an atypical way in
astronomy. This package is intended to provide such functionality, but
*not* to replace `scipy.stats` if its implementation satisfies
astronomers' needs.

"""

from . import bayesian_blocks as _bb
from . import (
    biweight,
    circstats,
    funcs,
    info_theory,
    jackknife,
    sigma_clipping,
    spatial,
)
from . import histogram as _hist
from .bayesian_blocks import *
from .biweight import *
from .circstats import *
from .funcs import *
from .histogram import *
from .info_theory import *
from .jackknife import *
from .sigma_clipping import *
from .spatial import *

__all__ = []
__all__ += funcs.__all__
__all__ += biweight.__all__
__all__ += sigma_clipping.__all__
__all__ += jackknife.__all__
__all__ += circstats.__all__
__all__ += _bb.__all__
__all__ += _hist.__all__
__all__ += info_theory.__all__
__all__ += spatial.__all__



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
import pytest

from astropy import convolution as conv
from astropy.tests.helper import check_pickling_recovery, pickle_protocol  # noqa: F401


@pytest.mark.parametrize(
    "original",
    [
        conv.CustomKernel(array=np.random.rand(15)),
        # FIXME: Gaussian1DKernel sometimes fails check_pickling_recovery()
        #        because Gaussian1D (modeling) sometimes fails it.
        #        Fixing the latter will automatically fix the former.
        # conv.Gaussian1DKernel(1.0, x_size=5),
        conv.Gaussian2DKernel(1.0, x_size=5, y_size=5),
    ],
    ids=lambda x: type(x).__name__,
)
def test_simple_object(pickle_protocol, original):  # noqa: F811
    # Tests easily instantiated objects
    check_pickling_recovery(original, pickle_protocol)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""|Cosmology| I/O, using |Cosmology.to_format| and |Cosmology.from_format|.

This module provides functions to transform a |Cosmology| object to and from another
|Cosmology| object. The functions are registered with ``convert_registry`` under the
format name "astropy.cosmology". You probably won't need to use these functions as they
are present mainly for completeness and testing.

    >>> from astropy.cosmology import Cosmology, Planck18
    >>> Planck18.to_format("astropy.cosmology") is Planck18
    True
    >>> Cosmology.from_format(Planck18) is Planck18
    True
"""

from astropy.cosmology._src.core import _COSMOLOGY_CLASSES, Cosmology
from astropy.cosmology._src.typing import _CosmoT
from astropy.cosmology.io import convert_registry

__all__: list[str] = []  # nothing is publicly scoped


def from_cosmology(
    cosmo: _CosmoT, /, cosmology: type[_CosmoT] | str | None = None, **kwargs: object
) -> _CosmoT:
    """Return the |Cosmology| unchanged.

    Parameters
    ----------
    cosmo : `~astropy.cosmology.Cosmology`
        The cosmology to return.
    cosmology : type[`~astropy.cosmology.Cosmology`] | str | None, optional
        The |Cosmology| class to check against. If not `None`, ``cosmo`` is checked
        for correctness.
    **kwargs
        This argument is required for compatibility with the standard set of
        keyword arguments in format |Cosmology.from_format|.

    Returns
    -------
    `~astropy.cosmology.Cosmology` subclass instance
        Just ``cosmo`` passed through.

    Raises
    ------
    TypeError
        If the |Cosmology| object is not an instance of ``cosmo`` (and
        ``cosmology`` is not `None`).

    Examples
    --------
    >>> from astropy.cosmology import Cosmology, Planck18
    >>> print(Cosmology.from_format(Planck18))
    FlatLambdaCDM(name="Planck18", H0=67.66 km / (Mpc s), Om0=0.30966,
                  Tcmb0=2.7255 K, Neff=3.046, m_nu=[0. 0. 0.06] eV, Ob0=0.04897)
    """
    # hot path
    if cosmology is None:
        return cosmo

    cosmo_cls = (
        _COSMOLOGY_CLASSES[cosmology] if isinstance(cosmology, str) else cosmology
    )
    if not isinstance(cosmo, cosmo_cls):
        raise TypeError(f"cosmology {cosmo} is not an {cosmo_cls} instance.")

    return cosmo


def to_cosmology(cosmo: _CosmoT, *args: object) -> _CosmoT:
    """Return the |Cosmology| unchanged.

    Parameters
    ----------
    cosmo : `~astropy.cosmology.Cosmology`
        The cosmology to return.
    *args : object
        Not used.

    Returns
    -------
    `~astropy.cosmology.Cosmology` subclass instance
        Just ``cosmo`` passed through.

    Examples
    --------
    >>> from astropy.cosmology import Planck18
    >>> Planck18.to_format("astropy.cosmology") is Planck18
    True
    """
    return cosmo


def cosmology_identify(
    origin: str, format: str | None, *args: object, **kwargs: object
) -> bool:
    """Identify if object is a `~astropy.cosmology.Cosmology`.

    This checks if the 2nd argument is a |Cosmology| instance and the format is
    "astropy.cosmology" or `None`.

    Returns
    -------
    bool
    """
    itis = False
    if origin == "read":
        itis = isinstance(args[1], Cosmology) and (
            format in (None, "astropy.cosmology")
        )
    return itis


# ===================================================================
# Register

convert_registry.register_reader("astropy.cosmology", Cosmology, from_cosmology)
convert_registry.register_writer("astropy.cosmology", Cosmology, to_cosmology)
convert_registry.register_identifier("astropy.cosmology", Cosmology, cosmology_identify)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module provides functions to help with testing against iraf tasks.
"""

import numpy as np

from astropy.logger import log


class IdentifyRecord:
    """
    Represents a database record for the onedspec.identify task.

    Attributes
    ----------
    recstr: string
            the record as a string
    fields: dict
            the fields in the record
    x: array
       the X values of the identified features
       this represents values on axis1 (image rows)
    y: int
       the Y values of the identified features
       (image columns)
    z: array
       the values which X maps into
    coeff: array
        function (modelname) coefficients
    """

    def __init__(self, recstr):
        self.recstr = recstr
        self.fields = self.get_fields()
        self._flatcoeff = self.fields["coefficients"].ravel()
        self.x = self.fields["features"][:, 0]
        self.y = self.get_ydata()
        self.z = self.fields["features"][:, 1]
        self.coeff = self._flatcoeff[4:]

    def aslist(self):
        reclist = self.recstr.split("\n")
        reclist = [entry.strip() for entry in reclist]
        [reclist.remove(entry) for entry in reclist if len(entry) == 0]
        return reclist

    def get_fields(self):
        # read record fields as an array
        fields = {}
        flist = self.aslist()
        numfields = len(flist)
        for i in range(numfields):
            line = flist[i]
            if line and line[0].isalpha():
                field = line.split()
                if i + 1 < numfields:
                    if not flist[i + 1][0].isalpha():
                        fields[field[0]] = self.read_array_field(
                            flist[i : i + int(field[1]) + 1]
                        )
                    else:
                        fields[field[0]] = " ".join(s for s in field[1:])
                else:
                    fields[field[0]] = " ".join(s for s in field[1:])
            else:
                continue
        return fields

    def read_array_field(self, fieldlist):
        # Turn an iraf record array field into a numpy array
        fieldline = [entry.split() for entry in fieldlist[1:]]
        # take only the first 3 columns
        # identify writes also strings at the end of some field lines
        xyz = [entry[:3] for entry in fieldline]
        try:
            farr = np.array(xyz)
        except Exception:
            log.debug(f"Could not read array field {fieldlist[0].split()[0]}")
        return farr.astype(np.float64)

    def get_range(self):
        low = self._flatcoeff[2]
        high = self._flatcoeff[3]
        return [low, high]

    def get_ydata(self):
        image = self.fields["image"]
        left = image.find("[") + 1
        right = image.find("]")
        section = image[left:right]
        if "," in section:
            yind = image.find(",") + 1
            return int(image[yind:-1])
        else:
            return int(section)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# Copied from astropy/convolution/setup_package.py

import os

from numpy import get_include as get_numpy_include
from setuptools import Extension

C_TIME_PKGDIR = os.path.relpath(os.path.dirname(__file__))

SRC_FILES = [
    os.path.join(C_TIME_PKGDIR, filename) for filename in ["src/parse_times.c"]
]


def get_extensions():
    # Add '-Rpass-missed=.*' to ``extra_compile_args`` when compiling with clang
    # to report missed optimizations
    _time_ext = Extension(
        name="astropy.time._parse_times",
        sources=SRC_FILES,
        include_dirs=[get_numpy_include()],
        language="c",
    )

    return [_time_ext]



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Testing :mod:`astropy.cosmology.FLRW` neutrinos."""

import numpy as np
import pytest

import astropy.units as u
from astropy.cosmology import FLRW, wCDM
from astropy.utils.compat.optional_deps import HAS_SCIPY

##############################################################################
# TYPES


class W1(FLRW):
    """
    This class is to test whether the routines work correctly if one only overloads w(z).
    """

    def __init__(self):
        super().__init__(70.0, 0.27, 0.73, Tcmb0=0.0, name="test_cos")
        self.__dict__["w0"] = -0.9

    def w(self, z):
        return self.w0 * np.ones_like(z)


class W1nu(FLRW):
    """Similar, but with neutrinos."""

    def __init__(self):
        super().__init__(
            70.0, 0.27, 0.73, Tcmb0=3.0, m_nu=0.1 * u.eV, name="test_cos_nu"
        )
        self.__dict__["w0"] = -0.8

    def w(self, z):
        return self.w0 * np.ones_like(z)


##############################################################################
# TESTS
##############################################################################


@pytest.mark.skipif(not HAS_SCIPY, reason="test requires scipy")
def test_de_subclass():
    z = [0.2, 0.4, 0.6, 0.9]

    # This is the comparison object
    cosmo = wCDM(H0=70, Om0=0.27, Ode0=0.73, w0=-0.9, Tcmb0=0.0)
    # Values taken from Ned Wrights advanced cosmo calculator, Aug 17 2012
    assert u.allclose(
        cosmo.luminosity_distance(z), [975.5, 2158.2, 3507.3, 5773.1] * u.Mpc, rtol=1e-3
    )

    # Now try the subclass that only gives w(z)
    cosmo = W1()
    assert u.allclose(
        cosmo.luminosity_distance(z), [975.5, 2158.2, 3507.3, 5773.1] * u.Mpc, rtol=1e-3
    )
    # Test efunc
    assert u.allclose(cosmo.efunc(1.0), 1.7489240754, rtol=1e-5)
    assert u.allclose(cosmo.efunc([0.5, 1.0]), [1.31744953, 1.7489240754], rtol=1e-5)
    assert u.allclose(cosmo.inv_efunc([0.5, 1.0]), [0.75904236, 0.57178011], rtol=1e-5)
    # Test de_density_scale
    assert u.allclose(cosmo.de_density_scale(1.0), 1.23114444, rtol=1e-4)
    assert u.allclose(
        cosmo.de_density_scale([0.5, 1.0]), [1.12934694, 1.23114444], rtol=1e-4
    )


@pytest.mark.skipif(not HAS_SCIPY, reason="test requires scipy")
def test_efunc_vs_invefunc_flrw():
    """Test that efunc and inv_efunc give inverse values"""
    z0 = 0.5
    z = np.array([0.5, 1.0, 2.0, 5.0])

    # FLRW is abstract, so requires W1 defined earlier
    # This requires scipy, unlike the built-ins, because it
    # calls de_density_scale, which has an integral in it
    cosmo = W1()
    assert u.allclose(cosmo.efunc(z0), 1.0 / cosmo.inv_efunc(z0))
    assert u.allclose(cosmo.efunc(z), 1.0 / cosmo.inv_efunc(z))
    # Add neutrinos
    cosmo = W1nu()
    assert u.allclose(cosmo.efunc(z0), 1.0 / cosmo.inv_efunc(z0))
    assert u.allclose(cosmo.efunc(z), 1.0 / cosmo.inv_efunc(z))



"""Static typing for :mod:`astropy.cosmology`. PRIVATE API."""
# Licensed under a 3-clause BSD style license - see LICENSE.rst

__all__ = ["CosmoMeta", "FArray", "_CosmoT"]

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeAlias, TypeVar

import numpy as np
from numpy.typing import NDArray

if TYPE_CHECKING:
    import astropy.cosmology

_CosmoT = TypeVar("_CosmoT", bound="astropy.cosmology.Cosmology")
"""Type variable for :class:`~astropy.cosmology.Cosmology` and subclasses."""

CosmoMeta: TypeAlias = Mapping[Any, Any]
"""Type alias for cosmology metadata."""

FArray: TypeAlias = NDArray[np.floating]
"""Type alias for numpy array of floating dtype."""



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import astropy.config as _config
from astropy.utils.compat import optional_deps

from .column import Column, ColumnInfo, MaskedColumn, StringTruncateWarning

__all__ = [
    "BST",
    "Column",
    "ColumnGroups",
    "ColumnInfo",
    "Conf",
    "JSViewer",
    "MaskedColumn",
    "NdarrayMixin",
    "PprintIncludeExclude",
    "QTable",
    "Row",
    "SCEngine",
    "SerializedColumn",
    "SortedArray",
    "StringTruncateWarning",
    "Table",
    "TableAttribute",
    "TableColumns",
    "TableFormatter",
    "TableGroups",
    "TableMergeError",
    "TableReplaceWarning",
    "conf",
    "conf",
    "connect",
    "dstack",
    "hstack",
    "join",
    "join_distance",
    "join_skycoord",
    "registry",
    "represent_mixins_as_columns",
    "setdiff",
    "unique",
    "vstack",
]


class Conf(_config.ConfigNamespace):
    """
    Configuration parameters for `astropy.table`.
    """

    auto_colname = _config.ConfigItem(
        "col{0}",
        "The template that determines the name of a column if it cannot be "
        "determined. Uses new-style (format method) string formatting.",
        aliases=["astropy.table.column.auto_colname"],
    )
    default_notebook_table_class = _config.ConfigItem(
        "table-striped table-bordered table-condensed",
        "The table class to be used in Jupyter notebooks when displaying "
        "tables (and not overridden). See <https://getbootstrap.com/css/#tables "
        "for a list of useful bootstrap classes.",
    )
    replace_warnings = _config.ConfigItem(
        [],
        "List of conditions for issuing a warning when replacing a table "
        "column using setitem, e.g. t['a'] = value.  Allowed options are "
        "'always', 'slice', 'refcount', 'attributes'.",
        "string_list",
    )
    replace_inplace = _config.ConfigItem(
        False,
        "Always use in-place update of a table column when using setitem, "
        "e.g. t['a'] = value.  This overrides the default behavior of "
        "replacing the column entirely with the new value when possible. "
        "This configuration option will be deprecated and then removed in "
        "subsequent major releases.",
    )


conf = Conf()

# Finally import the formats for the read and write method but delay building
# the documentation until all are loaded. (#5275)
from astropy.io import registry

from . import connect
from .bst import BST
from .groups import ColumnGroups, TableGroups
from .operations import (
    TableMergeError,
    dstack,
    hstack,
    join,
    join_distance,
    join_skycoord,
    setdiff,
    unique,
    vstack,
)
from .serialize import SerializedColumn, represent_mixins_as_columns
from .soco import SCEngine
from .sorted_array import SortedArray
from .table import (
    NdarrayMixin,
    PprintIncludeExclude,
    QTable,
    Row,
    Table,
    TableAttribute,
    TableColumns,
    TableFormatter,
    TableReplaceWarning,
)

with registry.delay_doc_updates(Table):
    # Import routines that connect readers/writers to astropy.table
    import astropy.io.ascii.connect
    import astropy.io.fits.connect
    import astropy.io.misc.connect
    import astropy.io.misc.pandas.connect
    import astropy.io.votable.connect

    from .jsviewer import JSViewer

    if optional_deps.HAS_ASDF_ASTROPY:
        import asdf_astropy.io.connect



from itertools import permutations

import numpy as np

__all__ = ["auto_assign_coord_positions"]


def auto_assign_coord_positions(ax):
    """
    Given a ``WCSAxes`` instance, automatically update any dynamic tick, tick
    label and axis label positions.

    This function operates in-place on the axes and assumes that
    ``_update_ticks`` has already been called on all the ``CoordinateHelper``
    instances.
    """
    # Since ticks, tick labels and axis labels can all be auto or fixed, we need
    # a few rules to decide in what order to process things:
    #
    # * We should deal with axis labels last - if they are set to auto they
    #   should be made to match the tick labels.
    # * If ticks and tick labels are both set to automatic, then we first
    #   process tick labels and at the end we can make ticks match tick labels.
    # * We should use existing fixed tick label assignments to exclude certain
    #   spines.
    # * Fixed tick positions should not exclude other labels from being placed
    #   on that same spine since multiple different ticks might appear on the
    #   same axis.
    # * If ticks are not auto and tick labels are auto then tick labels should be
    #   placed at a spine containing ticks.

    # Start off by simplifying some cases that don't require algorithmic
    # positioning. First, if tick labels are shown at fixed positions, we
    # should just adjust any auto ticks and axis labels to match.
    for coords in ax._all_coords:
        for coord in coords:
            if "#" not in (pos := coord.get_ticklabel_position()):
                if "#" in coord.get_ticks_position():
                    coord.set_ticks_position(pos + ["#"])
                if "#" in coord.get_axislabel_position():
                    coord.set_axislabel_position(pos + ["#"])

    # At this point, all coordinates requiring automatic placement have
    # tick labels requiring automatic placement (any coordinates with fixed
    # tick label placement are dealt with). We construct a list of these
    # coordinates and also keep track of spines on which tick labels are
    # already fixed.

    auto_coords = []
    already_used = []
    for coords in ax._all_coords:
        for coord in coords:
            pos = coord.get_ticklabel_position()
            if "#" in pos:
                auto_coords.append(coord)
            else:
                already_used += list(pos)

    # If there are no more coordinates with auto settings, we are all done
    if len(auto_coords) == 0:
        return

    # Extract the spines for the frame
    spines = coords.frame._spine_auto_position_order

    # Construct a new list of spines taking into account excluded ones
    spines = "".join(s for s in spines if s not in already_used)

    # We create an iterable of different assignments of spines to
    # coords, where empty string means the coordinate will not be shown
    # on any axis.
    if len(auto_coords) > len(spines):
        spines = spines + "".join(" " * (len(auto_coords) - len(spines)))
    options = permutations(spines, r=len(auto_coords))

    # We now loop through and check different assignments based on the options
    # and try and find the one that maximises the number of tick labels in the
    # overall plot.
    n_tick_max = -1
    best_option = None
    for option in options:
        # Check if option is consistent with any fixed tick positions - that
        # is, if ticks were explicitly requested on axes say b and t then we
        # shouldn't automatically put tick labels on say l or r.
        consistent = True
        for c, s in zip(auto_coords, option):
            pos = c.get_ticks_position()
            if "#" not in pos and s not in pos:
                consistent = False
                break
        if not consistent:
            continue

        # Determine the number of tick labels on each axis
        n_on_each = {s: len(c._ticks.world[s]) for c, s in zip(auto_coords, option)}

        # Determine the total number of tick labels
        n_tick = sum(n_on_each.values())

        # Determine a sorted version of this list by spine axis order
        n_on_each_sorted = [n_on_each.get(s, 0) for s in spines]

        # We should ideally not have empty spines, so if there are any non-zero
        # values in n_on_each then we should add a penalty for every 0 value
        # that occurs before the last non-zero value
        idx = np.nonzero(n_on_each_sorted)[0]
        if len(idx) > 0:
            n_tick -= n_on_each_sorted[: idx[-1]].count(0)

        # Keep track of the best option so far
        if n_tick > n_tick_max:
            n_tick_max = n_tick
            best_option = option

    # Finalize assignments
    for coord, spine in zip(auto_coords, best_option):
        position = [spine, "#"] if spine != " " else "#"
        if "#" in coord.get_ticks_position():
            coord.set_ticks_position(position)
        if "#" in coord.get_ticklabel_position():
            coord.set_ticklabel_position(position)
        if "#" in coord.get_axislabel_position():
            coord.set_axislabel_position(position)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Scale factor.

This is private API. See `~astropy.cosmology.traits` for public API.

"""

__all__ = ["ScaleFactor"]


from numpy.typing import ArrayLike

import astropy.units as u
from astropy.cosmology._src.utils import aszarr, deprecated_keywords


class ScaleFactor:
    """The trait for computing the cosmological scale factor.

    The scale factor is defined as :math:`a = a_0 / (1 + z)`.

    """

    @property
    def scale_factor0(self) -> u.Quantity:
        r"""Scale factor at redshift 0.

        The scale factor is defined as :math:`a = a_0 / (1 + z)`. The common convention
        is to set :math:`a_0 = 1`. However, in some cases, like in some old CMB papers,
        :math:`a_0` is used to normalize `a` to be a convenient number at the redshift
        of interest for that paper. Explicitly using :math:`a_0` in both calculation and
        code avoids ambiguity.
        """
        return 1 << u.one

    @deprecated_keywords("z", since="7.0")
    def scale_factor(self, z: u.Quantity | ArrayLike) -> u.Quantity:
        """Compute the scale factor at redshift ``z``.

        The scale factor is defined as :math:`a = a_0 / (1 + z)`.

        Parameters
        ----------
        z : Quantity-like ['redshift'] | array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        |Quantity|
            Scale factor at each input redshift.
        """
        return self.scale_factor0 / (aszarr(z) + 1)



# Licensed under a 3-clause BSD style license - see LICENSE.rst


import copy
import pickle
from io import StringIO

import numpy as np
import pytest

from astropy import coordinates, time
from astropy import units as u
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.coordinates.tests.helper import skycoord_equal
from astropy.coordinates.tests.test_representation import representation_equal
from astropy.io.ascii.connect import _get_connectors_table
from astropy.table import (
    Column,
    NdarrayMixin,
    QTable,
    Table,
    hstack,
    join,
    serialize,
    table_helpers,
    vstack,
)
from astropy.table.column import BaseColumn
from astropy.table.serialize import represent_mixins_as_columns
from astropy.table.table_helpers import ArrayWrapper
from astropy.utils.compat import NUMPY_LT_2_0
from astropy.utils.data_info import ParentDtypeInfo
from astropy.utils.exceptions import AstropyUserWarning
from astropy.utils.metadata import MergeConflictWarning

from .conftest import MIXIN_COLS


def test_attributes(mixin_cols):
    """
    Required attributes for a column can be set.
    """
    m = mixin_cols["m"]
    m.info.name = "a"
    assert m.info.name == "a"

    m.info.description = "a"
    assert m.info.description == "a"

    # Cannot set unit for these classes
    if isinstance(
        m,
        (
            u.Quantity,
            coordinates.SkyCoord,
            time.Time,
            time.TimeDelta,
            coordinates.BaseRepresentationOrDifferential,
            coordinates.StokesCoord,
        ),
    ):
        with pytest.raises(AttributeError):
            m.info.unit = u.m
    else:
        m.info.unit = u.m
        assert m.info.unit is u.m

    m.info.format = "a"
    assert m.info.format == "a"

    m.info.meta = {"a": 1}
    assert m.info.meta == {"a": 1}

    with pytest.raises(AttributeError):
        m.info.bad_attr = 1

    with pytest.raises(AttributeError):
        m.info.bad_attr


def check_mixin_type(table, table_col, in_col):
    # We check for QuantityInfo rather than just isinstance(col, u.Quantity)
    # since we want to treat EarthLocation as a mixin, even though it is
    # a Quantity subclass.
    if (
        isinstance(in_col.info, u.QuantityInfo) and type(table) is not QTable
    ) or isinstance(in_col, Column):
        assert type(table_col) is table.ColumnClass
    else:
        assert type(table_col) is type(in_col)

    # Make sure in_col got copied and creating table did not touch it
    assert in_col.info.name is None


def test_make_table(table_types, mixin_cols):
    """
    Make a table with the columns in mixin_cols, which is an ordered dict of
    three cols: 'a' and 'b' are table_types.Column type, and 'm' is a mixin.
    """
    t = table_types.Table(mixin_cols)
    check_mixin_type(t, t["m"], mixin_cols["m"])

    cols = list(mixin_cols.values())
    t = table_types.Table(cols, names=("i", "a", "b", "m"))
    check_mixin_type(t, t["m"], mixin_cols["m"])

    t = table_types.Table(cols)
    check_mixin_type(t, t["col3"], mixin_cols["m"])


def test_io_ascii_write():
    """
    Test that table with mixin column can be written by io.ascii for
    every pure Python writer.  No validation of the output is done,
    this just confirms no exceptions.
    """
    t = QTable(MIXIN_COLS)
    for fmt in _get_connectors_table():
        if (
            fmt["Write"]
            and ".fast_" not in fmt["Format"]
            and fmt["Format"] != "ascii.tdat"
        ):
            out = StringIO()
            t.write(out, format=fmt["Format"])


def test_votable_quantity_write(tmp_path):
    """
    Test that table with Quantity mixin column can be round-tripped by
    io.votable.  Note that FITS and HDF5 mixin support are tested (much more
    thoroughly) in their respective subpackage tests
    (io/fits/tests/test_connect.py and io/misc/tests/test_hdf5.py).
    """
    t = QTable()
    t["a"] = u.Quantity([1, 2, 4], unit="nm")

    filename = tmp_path / "table-tmp"
    t.write(filename, format="votable", overwrite=True)
    qt = QTable.read(filename, format="votable")
    assert isinstance(qt["a"], u.Quantity)
    assert qt["a"].unit == "nm"


@pytest.mark.remote_data
@pytest.mark.parametrize("table_types", (Table, QTable))
def test_io_time_write_fits_standard(tmp_path, table_types):
    """
    Test that table with Time mixin columns can be written by io.fits.
    Validation of the output is done. Test that io.fits writes a table
    containing Time mixin columns that can be partially round-tripped
    (metadata scale, location).

    Note that we postpone checking the "local" scale, since that cannot
    be done with format 'cxcsec', as it requires an epoch.
    """
    t = table_types([[1, 2], ["string", "column"]])
    for scale in time.STANDARD_TIME_SCALES:
        t["a" + scale] = time.Time(
            [[1, 2], [3, 4]],
            format="cxcsec",
            scale=scale,
            location=EarthLocation(-2446354, 4237210, 4077985, unit="m"),
        )
        t["b" + scale] = time.Time(
            ["1999-01-01T00:00:00.123456789", "2010-01-01T00:00:00"], scale=scale
        )
    t["c"] = [3.0, 4.0]

    filename = tmp_path / "table-tmp"

    # Show that FITS format succeeds
    with pytest.warns(AstropyUserWarning) as record:
        t.write(filename, format="fits", overwrite=True)

    # The exact sequence probably
    # does not matter too much, so we'll just try to match the *set* of warnings
    warnings = {wm.message.args[0] for wm in record}

    expected = {
        (
            'Time Column "btai" has no specified location, '
            "but global Time Position is present, "
            "which will be the default for this column in FITS specification."
        ),
        (
            'Time Column "btdb" has no specified location, '
            "but global Time Position is present, "
            "which will be the default for this column in FITS specification."
        ),
        (
            'Time Column "btcg" has no specified location, '
            "but global Time Position is present, "
            "which will be the default for this column in FITS specification."
        ),
        (
            'Time Column "btt" has no specified location, '
            "but global Time Position is present, which will be the default "
            "for this column in FITS specification."
        ),
        (
            'Time Column "butc" has no specified location, '
            "but global Time Position is present, "
            "which will be the default for this column in FITS specification."
        ),
        (
            'Earth Location "TOPOCENTER" for Time Column "atdb"'
            ' is incompatible with scale "TDB".'
        ),
        (
            'Earth Location "TOPOCENTER" for Time Column "atcb"'
            ' is incompatible with scale "TCB".'
        ),
        (
            'Time Column "but1" has no specified location, '
            "but global Time Position is present, "
            "which will be the default for this column in FITS specification."
        ),
        (
            'Time Column "btcb" has no specified location, '
            "but global Time Position is present, "
            "which will be the default for this column in FITS specification."
        ),
    }
    assert warnings == expected, f"Got some unexpected warnings\n{warnings - expected}"
    with pytest.warns(
        AstropyUserWarning,
        match='Time column reference position "TRPOSn" is not specified',
    ):
        tm = table_types.read(filename, format="fits", astropy_native=True)

    for scale in time.STANDARD_TIME_SCALES:
        for ab in ("a", "b"):
            name = ab + scale

            # Assert that the time columns are read as Time
            assert isinstance(tm[name], time.Time)

            # Assert that the scales round-trip
            assert tm[name].scale == t[name].scale

            # Assert that the format is jd
            assert tm[name].format == "jd"

            # Assert that the location round-trips
            assert tm[name].location == t[name].location

            # Finally assert that the column data round-trips
            assert (tm[name] == t[name]).all()

    for name in ("col0", "col1", "c"):
        # Assert that the non-time columns are read as Column
        assert isinstance(tm[name], Column)

        # Assert that the non-time columns' data round-trips
        assert (tm[name] == t[name]).all()

    # Test for conversion of time data to its value, as defined by its format
    for scale in time.STANDARD_TIME_SCALES:
        for ab in ("a", "b"):
            name = ab + scale
            t[name].info.serialize_method["fits"] = "formatted_value"

    t.write(filename, format="fits", overwrite=True)
    tm = table_types.read(filename, format="fits")

    for scale in time.STANDARD_TIME_SCALES:
        for ab in ("a", "b"):
            name = ab + scale

            assert not isinstance(tm[name], time.Time)
            assert (tm[name] == t[name].value).all()


@pytest.mark.parametrize("table_types", (Table, QTable))
def test_io_time_write_fits_local(tmp_path, table_types):
    """
    Test that table with a Time mixin with scale local can also be written
    by io.fits. Like ``test_io_time_write_fits_standard`` above, but avoiding
    ``cxcsec`` format, which requires an epoch and thus cannot be used for a
    local time scale.
    """
    t = table_types([[1, 2], ["string", "column"]])
    t["a_local"] = time.Time(
        [[50001, 50002], [50003, 50004]],
        format="mjd",
        scale="local",
        location=EarthLocation(-2446354, 4237210, 4077985, unit="m"),
    )
    t["b_local"] = time.Time(
        ["1999-01-01T00:00:00.123456789", "2010-01-01T00:00:00"], scale="local"
    )
    t["c"] = [3.0, 4.0]

    filename = tmp_path / "table-tmp"

    # Show that FITS format succeeds

    with pytest.warns(
        AstropyUserWarning, match='Time Column "b_local" has no specified location'
    ):
        t.write(filename, format="fits", overwrite=True)

    with pytest.warns(
        AstropyUserWarning,
        match='Time column reference position "TRPOSn" is not specified.',
    ):
        tm = table_types.read(filename, format="fits", astropy_native=True)

    for ab in ("a", "b"):
        name = ab + "_local"

        # Assert that the time columns are read as Time
        assert isinstance(tm[name], time.Time)

        # Assert that the scales round-trip
        assert tm[name].scale == t[name].scale

        # Assert that the format is jd
        assert tm[name].format == "jd"

        # Assert that the location round-trips
        assert tm[name].location == t[name].location

        # Finally assert that the column data round-trips
        assert (tm[name] == t[name]).all()

    for name in ("col0", "col1", "c"):
        # Assert that the non-time columns are read as Column
        assert isinstance(tm[name], Column)

        # Assert that the non-time columns' data round-trips
        assert (tm[name] == t[name]).all()

    # Test for conversion of time data to its value, as defined by its format.
    for ab in ("a", "b"):
        name = ab + "_local"
        t[name].info.serialize_method["fits"] = "formatted_value"

    t.write(filename, format="fits", overwrite=True)
    tm = table_types.read(filename, format="fits")

    for ab in ("a", "b"):
        name = ab + "_local"

        assert not isinstance(tm[name], time.Time)
        assert (tm[name] == t[name].value).all()


def test_votable_mixin_write_fail(mixin_cols):
    """
    Test that table with mixin columns (excluding Quantity) cannot be written by
    io.votable.
    """
    t = QTable(mixin_cols)
    # Only do this test if there are unsupported column types (i.e. anything besides
    # BaseColumn and Quantity class instances).
    unsupported_cols = t.columns.not_isinstance((BaseColumn, u.Quantity))

    if not unsupported_cols:
        pytest.skip("no unsupported column types")

    out = StringIO()
    with pytest.raises(ValueError) as err:
        t.write(out, format="votable")
    assert "cannot write table with mixin column(s)" in str(err.value)


def test_join(table_types):
    """
    Join tables with mixin cols.  Use column "i" as proxy for what the
    result should be for each mixin.
    """
    t1 = table_types.Table()
    t1["a"] = table_types.Column(["a", "b", "b", "c"])
    t1["i"] = table_types.Column([0, 1, 2, 3])
    for name, col in MIXIN_COLS.items():
        t1[name] = col

    t2 = table_types.Table(t1)
    t2["a"] = ["b", "c", "a", "d"]

    for name in MIXIN_COLS:
        t1[name].info.description = name
        t2[name].info.description = name + "2"

    for join_type in ("inner", "left"):
        t12 = join(t1, t2, keys="a", join_type=join_type)
        idx1 = t12["i_1"]
        idx2 = t12["i_2"]
        for name, col in MIXIN_COLS.items():
            name1 = name + "_1"
            name2 = name + "_2"
            assert_table_name_col_equal(t12, name1, col[idx1])
            assert_table_name_col_equal(t12, name2, col[idx2])
            assert t12[name1].info.description == name
            assert t12[name2].info.description == name + "2"

    for join_type in ("outer", "right"):
        with pytest.raises(NotImplementedError) as exc:
            t12 = join(t1, t2, keys="a", join_type=join_type)
        assert "join requires masking column" in str(exc.value)

    with pytest.raises(TypeError) as exc:
        t12 = join(t1, t2, keys=["a", "skycoord"])
    assert "one or more key columns are not sortable" in str(exc.value)

    # Join does work for a mixin which is a subclass of np.ndarray
    with pytest.warns(
        MergeConflictWarning,
        match="In merged column 'quantity' the 'description' attribute does not match",
    ):
        t12 = join(t1, t2, keys=["quantity"])
    assert np.all(t12["a_1"] == t1["a"])


def test_hstack(table_types):
    """
    Hstack tables with mixin cols.  Use column "i" as proxy for what the
    result should be for each mixin.
    """
    t1 = table_types.Table()
    t1["i"] = table_types.Column([0, 1, 2, 3])
    for name, col in MIXIN_COLS.items():
        t1[name] = col
        t1[name].info.description = name
        t1[name].info.meta = {"a": 1}

    for join_type in ("inner", "outer"):
        for chop in (True, False):
            t2 = table_types.Table(t1)
            if chop:
                t2 = t2[:-1]
                if join_type == "outer":
                    with pytest.raises(NotImplementedError) as exc:
                        t12 = hstack([t1, t2], join_type=join_type)
                    assert "hstack requires masking column" in str(exc.value)
                    continue

            t12 = hstack([t1, t2], join_type=join_type)
            idx1 = t12["i_1"]
            idx2 = t12["i_2"]
            for name, col in MIXIN_COLS.items():
                name1 = name + "_1"
                name2 = name + "_2"
                assert_table_name_col_equal(t12, name1, col[idx1])
                assert_table_name_col_equal(t12, name2, col[idx2])
                for attr in ("description", "meta"):
                    assert getattr(t1[name].info, attr) == getattr(
                        t12[name1].info, attr
                    )
                    assert getattr(t2[name].info, attr) == getattr(
                        t12[name2].info, attr
                    )


def assert_table_name_col_equal(t, name, col):
    """
    Assert all(t[name] == col), with special handling for known mixin cols.
    """
    __tracebackhide__ = True
    if isinstance(col, coordinates.SkyCoord):
        assert np.all(t[name].ra == col.ra)
        assert np.all(t[name].dec == col.dec)
    elif isinstance(col, coordinates.BaseRepresentationOrDifferential):
        assert np.all(representation_equal(t[name], col))
    elif isinstance(col, u.Quantity):
        if type(t) is QTable:
            assert np.all(t[name] == col)
    elif isinstance(col, table_helpers.ArrayWrapper):
        assert np.all(t[name].data == col.data)
    else:
        assert np.all(t[name] == col)


def test_get_items(mixin_cols):
    """
    Test that slicing / indexing table gives right values and col attrs inherit
    """
    attrs = ("name", "unit", "dtype", "format", "description", "meta")
    m = mixin_cols["m"]
    m.info.name = "m"
    m.info.format = "{0}"
    m.info.description = "d"
    m.info.meta = {"a": 1}
    t = QTable([m])
    for item in ([1, 3], np.array([0, 2]), slice(1, 3)):
        t2 = t[item]
        m2 = m[item]
        assert_table_name_col_equal(t2, "m", m[item])
        for attr in attrs:
            assert getattr(t2["m"].info, attr) == getattr(m.info, attr)
            assert getattr(m2.info, attr) == getattr(m.info, attr)


def test_info_preserved_pickle_copy_init(mixin_cols):
    """
    Test copy, pickle, and init from class roundtrip preserve info.  This
    tests not only the mixin classes but a regular column as well.
    """

    def pickle_roundtrip(c):
        # protocol=5 matches the default for Python 3.14 and later
        # and is needed to preserve byteorder (available since Python 3.8)
        return pickle.loads(pickle.dumps(c, protocol=5))

    def init_from_class(c):
        return c.__class__(c)

    attrs = ("name", "unit", "dtype", "format", "description", "meta")
    for colname in ("i", "m"):
        m = mixin_cols[colname]
        m.info.name = colname
        m.info.format = "{0}"
        m.info.description = "d"
        m.info.meta = {"a": 1}
        for func in (copy.copy, copy.deepcopy, pickle_roundtrip, init_from_class):
            m2 = func(m)
            for attr in attrs:
                assert getattr(m2.info, attr) == getattr(m.info, attr)


def check_share_memory(col1, col2, copy):
    """Check whether data attributes in col1 and col2 share memory.

    If copy=True, this should not be the case for any, while
    if copy=False, all should share memory.
    """
    if isinstance(col1, SkyCoord):
        # For SkyCoord, .info does not access actual data by default,
        # but rather attributes like .ra, which are copies.
        map1 = col1.data.info._represent_as_dict()
        map2 = col2.data.info._represent_as_dict()
    else:
        map1 = col1.info._represent_as_dict()
        map2 = col2.info._represent_as_dict()

    # Check array attributes only (in principle, could iterate on, e.g.,
    # differentials in representations, but this is enough for table).
    shared = [
        np.may_share_memory(v1, v2)
        for (v1, v2) in zip(map1.values(), map2.values())
        if isinstance(v1, np.ndarray) and v1.shape
    ]
    if copy:
        assert not any(shared)
    else:
        assert all(shared)


@pytest.mark.parametrize("copy", [True, False])
def test_add_column(mixin_cols, copy):
    """
    Test that adding a column preserves values and attributes.
    For copy=True, the data should be independent;
    for copy=False, the data should be shared, but the instance independent.
    """
    attrs = ("name", "unit", "dtype", "format", "description", "meta")
    m = mixin_cols["m"]
    assert m.info.name is None

    # Make sure adding column in various ways doesn't touch info.
    t = QTable([m], names=["a"], copy=copy)
    assert m.info.name is None
    check_share_memory(m, t["a"], copy=copy)

    t["new"] = m
    assert m.info.name is None
    check_share_memory(m, t["new"], copy=True)

    m.info.name = "m"
    m.info.format = "{0}"
    m.info.description = "d"
    m.info.meta = {"a": 1}
    t = QTable([m], copy=copy)
    assert t.colnames == ["m"]
    check_share_memory(m, t["m"], copy=copy)

    t = QTable([m], names=["m1"], copy=copy)
    assert m.info.name == "m"
    assert t.colnames == ["m1"]
    check_share_memory(m, t["m1"], copy=copy)

    # Add columns m2, m3, m4 by two different methods and test expected equality
    t["m2"] = m
    check_share_memory(m, t["m2"], copy=True)
    m.info.name = "m3"
    t.add_columns([m], copy=copy)
    check_share_memory(m, t["m3"], copy=copy)
    for name in ("m2", "m3"):
        assert_table_name_col_equal(t, name, m)
        for attr in attrs:
            if attr != "name":
                assert getattr(t["m1"].info, attr) == getattr(t[name].info, attr)
    # Also check that one can set using a scalar.
    s = m[0]
    if type(s) is type(m) and "info" in s.__dict__:
        # We're not going to worry about testing classes for which scalars
        # are a different class than the real array, or where info is not copied.
        t["s"] = m[0]
        assert_table_name_col_equal(t, "s", m[0])
        check_share_memory(m, t["s"], copy=True)
        for attr in attrs:
            if attr != "name":
                assert getattr(t["m1"].info, attr) == getattr(t["s"].info, attr)

    # While we're add it, also check a length-1 table.
    t = QTable([m[1:2]], names=["m"], copy=copy)
    check_share_memory(m, t["m"], copy=copy)
    if type(s) is type(m) and "info" in s.__dict__:
        t["s"] = m[0]
        assert_table_name_col_equal(t, "s", m[0])
        for attr in attrs:
            if attr != "name":
                assert getattr(t["m1"].info, attr) == getattr(t["s"].info, attr)


def test_vstack():
    """
    Vstack tables with mixin cols.
    """
    t1 = QTable(MIXIN_COLS)
    t2 = QTable(MIXIN_COLS)
    with pytest.raises(NotImplementedError):
        vstack([t1, t2])


@pytest.mark.parametrize("empty_table", [False, True])
def test_insert_row(mixin_cols, empty_table):
    """
    Test inserting a row, which works for Column, Quantity, Time and SkyCoord.
    """
    t0 = QTable(mixin_cols)
    if empty_table:
        t = t0[:0].copy()
        insert_index = 0
        result_indices = [-1]
    else:
        t = t0.copy()
        insert_index = 1
        result_indices = [0, -1, 1, 2, 3]
    t["m"].info.description = "d"
    if isinstance(
        t["m"], (u.Quantity, Column, time.Time, time.TimeDelta, coordinates.SkyCoord)
    ):
        t.insert_row(insert_index, t0[-1])

        for name in t.colnames:
            assert np.all(t[name] == t0[name][result_indices])

        assert np.all(t == t0[result_indices])
        assert t["m"].info.description == "d"
    else:
        with pytest.raises(ValueError) as exc:
            t.insert_row(insert_index, t0[-1])
        assert "Unable to insert row" in str(exc.value)


def test_insert_row_bad_unit():
    """
    Insert a row into a QTable with the wrong unit
    """
    t = QTable([[1] * u.m])
    with pytest.raises(ValueError) as exc:
        t.insert_row(0, (2 * u.m / u.s,))
    assert "'m / s' (speed/velocity) and 'm' (length) are not convertible" in str(
        exc.value
    )


def test_convert_np_array(mixin_cols):
    """
    Test that converting to numpy array creates an object dtype and that
    each instance in the array has the expected type.
    """
    t = QTable(mixin_cols)
    ta = t.as_array()
    m = mixin_cols["m"]
    dtype_kind = m.dtype.kind if hasattr(m, "dtype") else "O"
    assert ta["m"].dtype.kind == dtype_kind


def test_assignment_and_copy():
    """
    Test that assignment of an int, slice, and fancy index works.
    Along the way test that copying table works.
    """
    for name in ("quantity", "arraywrap"):
        m = MIXIN_COLS[name]
        t0 = QTable([m], names=["m"])
        for i0, i1 in (
            (1, 2),
            (slice(0, 2), slice(1, 3)),
            (np.array([1, 2]), np.array([2, 3])),
        ):
            t = t0.copy()
            t["m"][i0] = m[i1]
            if name == "arraywrap":
                assert np.all(t["m"].data[i0] == m.data[i1])
                assert np.all(t0["m"].data[i0] == m.data[i0])
                assert np.all(t0["m"].data[i0] != t["m"].data[i0])
            else:
                assert np.all(t["m"][i0] == m[i1])
                assert np.all(t0["m"][i0] == m[i0])
                assert np.all(t0["m"][i0] != t["m"][i0])


def test_conversion_qtable_table():
    """
    Test that a table round trips from QTable => Table => QTable
    """
    qt = QTable(MIXIN_COLS)
    names = qt.colnames
    for name in names:
        qt[name].info.description = name

    t = Table(qt)
    for name in names:
        assert t[name].info.description == name
        if name == "quantity":
            assert np.all(t["quantity"] == qt["quantity"].value)
            assert np.all(t["quantity"].unit is qt["quantity"].unit)
            assert isinstance(t["quantity"], t.ColumnClass)
        else:
            assert_table_name_col_equal(t, name, qt[name])

    qt2 = QTable(qt)
    for name in names:
        assert qt2[name].info.description == name
        assert_table_name_col_equal(qt2, name, qt[name])


def test_setitem_as_column_name():
    """
    Test for mixin-related regression described in #3321.
    """
    t = Table()
    t["a"] = ["x", "y"]
    t["b"] = "b"  # Previously was failing with KeyError
    assert np.all(t["a"] == ["x", "y"])
    assert np.all(t["b"] == ["b", "b"])


def test_quantity_representation():
    """
    Test that table representation of quantities does not have unit
    """
    t = QTable([[1, 2] * u.m])
    assert t.pformat() == [
        "col0",
        " m  ",
        "----",
        " 1.0",
        " 2.0",
    ]


@pytest.mark.parametrize(
    "c, expected_pformat",
    [
        pytest.param(
            coordinates.CartesianRepresentation([0], [1], [0], unit=u.one),
            # With no unit we get "None" in the unit row
            [
                "    col0    ",
                "------------",
                "(0., 1., 0.)",
            ]
            if NUMPY_LT_2_0
            else [
                "      col0     ",
                "---------------",
                "(0.0, 1.0, 0.0)",
            ],
            id="cartesian_wo_unit",
        ),
        pytest.param(
            coordinates.CartesianRepresentation([0], [1], [0], unit="m"),
            [
                "    col0    ",
                "     m      ",
                "------------",
                "(0., 1., 0.)",
            ]
            if NUMPY_LT_2_0
            else [
                "      col0     ",
                "       m       ",
                "---------------",
                "(0.0, 1.0, 0.0)",
            ],
            id="cartesian_w_unit",
        ),
        pytest.param(
            coordinates.SphericalRepresentation([10] * u.deg, [20] * u.deg, [1] * u.pc),
            [
                "     col0     ",
                " deg, deg, pc ",
                "--------------",
                "(10., 20., 1.)",
            ]
            if NUMPY_LT_2_0
            else [
                "       col0      ",
                "   deg, deg, pc  ",
                "-----------------",
                "(10.0, 20.0, 1.0)",
            ],
            id="spherical",
        ),
        pytest.param(
            coordinates.UnitSphericalRepresentation([10] * u.deg, [20] * u.deg),
            [
                "   col0   ",
                "   deg    ",
                "----------",
                "(10., 20.)",
            ]
            if NUMPY_LT_2_0
            else [
                "    col0    ",
                "    deg     ",
                "------------",
                "(10.0, 20.0)",
            ],
            id="unitspherical",
        ),
        pytest.param(
            coordinates.SphericalCosLatDifferential(
                [10] * u.mas / u.yr, [2] * u.mas / u.yr, [10] * u.km / u.s
            ),
            [
                "           col0           ",
                "mas / yr, mas / yr, km / s",
                "--------------------------",
                "            (10., 2., 10.)",
            ]
            if NUMPY_LT_2_0
            else [
                "           col0           ",
                "mas / yr, mas / yr, km / s",
                "--------------------------",
                "         (10.0, 2.0, 10.0)",
            ],
            id="sphericalcoslatdifferential",
        ),
    ],
)
def test_representation_representation(c, expected_pformat):
    """
    Test that Representations are represented correctly.
    """
    t = Table([c])
    assert t.pformat() == expected_pformat


def test_skycoord_representation():
    """
    Test that skycoord representation works, both in the way that the
    values are output and in changing the frame representation.
    """
    # With no unit we get "None" in the unit row
    c = coordinates.SkyCoord([0], [1], [0], representation_type="cartesian")
    t = Table([c])
    assert t.pformat() == [
        "     col0     ",
        "None,None,None",
        "--------------",
        "   0.0,1.0,0.0",
    ]

    # Test that info works with a dynamically changed representation
    c = coordinates.SkyCoord([0], [1], [0], unit="m", representation_type="cartesian")
    t = Table([c])
    assert t.pformat() == [
        "    col0   ",
        "   m,m,m   ",
        "-----------",
        "0.0,1.0,0.0",
    ]

    t["col0"].representation_type = "unitspherical"
    assert t.pformat() == [
        "  col0  ",
        "deg,deg ",
        "--------",
        "90.0,0.0",
    ]

    t["col0"].representation_type = "cylindrical"
    assert t.pformat() == [
        "    col0    ",
        "  m,deg,m   ",
        "------------",
        "1.0,90.0,0.0",
    ]


@pytest.mark.parametrize("as_ndarray_mixin", [True, False])
def test_ndarray_mixin(as_ndarray_mixin):
    """
    Test directly adding various forms of structured ndarray columns to a table.
    Adding as NdarrayMixin is expected to be somewhat unusual after #12644
    (which provides full support for structured array Column's). This test shows
    that the end behavior is the same in both cases.
    """
    a = np.array([(1, "a"), (2, "b"), (3, "c"), (4, "d")], dtype="<i4,|U1")
    b = np.array(
        [(10, "aa"), (20, "bb"), (30, "cc"), (40, "dd")],
        dtype=[("x", "i4"), ("y", "U2")],
    )
    c = np.rec.fromrecords(
        [(100.0, "raa"), (200.0, "rbb"), (300.0, "rcc"), (400.0, "rdd")],
        names=["rx", "ry"],
    )
    d = np.arange(8, dtype="i8").reshape(4, 2)

    if as_ndarray_mixin:
        a = a.view(NdarrayMixin)
        b = b.view(NdarrayMixin)
        c = c.view(NdarrayMixin)
        d = d.view(NdarrayMixin)
        class_exp = NdarrayMixin
    else:
        class_exp = Column

    # Add one during initialization and the next as a new column.
    t = Table([a], names=["a"])
    t["b"] = b
    t["c"] = c
    t["d"] = d

    assert isinstance(t["a"], class_exp)

    assert t["a"][1][1] == a[1][1]
    assert t["a"][2][0] == a[2][0]

    assert t[1]["a"][1] == a[1][1]
    assert t[2]["a"][0] == a[2][0]

    assert isinstance(t["b"], class_exp)

    assert t["b"][1]["x"] == b[1]["x"]
    assert t["b"][1]["y"] == b[1]["y"]

    assert t[1]["b"]["x"] == b[1]["x"]
    assert t[1]["b"]["y"] == b[1]["y"]

    assert isinstance(t["c"], class_exp)

    assert t["c"][1]["rx"] == c[1]["rx"]
    assert t["c"][1]["ry"] == c[1]["ry"]

    assert t[1]["c"]["rx"] == c[1]["rx"]
    assert t[1]["c"]["ry"] == c[1]["ry"]

    assert isinstance(t["d"], class_exp)

    assert t["d"][1][0] == d[1][0]
    assert t["d"][1][1] == d[1][1]

    assert t[1]["d"][0] == d[1][0]
    assert t[1]["d"][1] == d[1][1]

    assert t.pformat(show_dtype=True) == (
        [
            "  a [f0, f1]     b [x, y]      c [rx, ry]      d    ",
            "(int32, str1) (int32, str2) (float64, str3) int64[2]",
            "------------- ------------- --------------- --------",
            "     (1, 'a')    (10, 'aa')   (100., 'raa')   0 .. 1",
            "     (2, 'b')    (20, 'bb')   (200., 'rbb')   2 .. 3",
            "     (3, 'c')    (30, 'cc')   (300., 'rcc')   4 .. 5",
            "     (4, 'd')    (40, 'dd')   (400., 'rdd')   6 .. 7",
        ]
        if NUMPY_LT_2_0
        else [
            "  a [f0, f1]     b [x, y]      c [rx, ry]      d    ",
            "(int32, str1) (int32, str2) (float64, str3) int64[2]",
            "------------- ------------- --------------- --------",
            "     (1, 'a')    (10, 'aa')  (100.0, 'raa')   0 .. 1",
            "     (2, 'b')    (20, 'bb')  (200.0, 'rbb')   2 .. 3",
            "     (3, 'c')    (30, 'cc')  (300.0, 'rcc')   4 .. 5",
            "     (4, 'd')    (40, 'dd')  (400.0, 'rdd')   6 .. 7",
        ]
    )


def test_possible_string_format_functions():
    """
    The QuantityInfo info class for Quantity implements a
    possible_string_format_functions() method that overrides the
    standard pprint._possible_string_format_functions() function.
    Test this.
    """
    t = QTable([[1, 2] * u.m])
    t["col0"].info.format = "%.3f"
    assert t.pformat() == [
        " col0",
        "  m  ",
        "-----",
        "1.000",
        "2.000",
    ]

    t["col0"].info.format = "hi {:.3f}"
    assert t.pformat() == [
        "  col0  ",
        "   m    ",
        "--------",
        "hi 1.000",
        "hi 2.000",
    ]

    t["col0"].info.format = ".4f"
    assert t.pformat() == [
        " col0 ",
        "  m   ",
        "------",
        "1.0000",
        "2.0000",
    ]


def test_rename_mixin_columns(mixin_cols):
    """
    Rename a mixin column.
    """
    t = QTable(mixin_cols)
    tc = t.copy()
    t.rename_column("m", "mm")
    assert t.colnames == ["i", "a", "b", "mm"]
    if isinstance(t["mm"], table_helpers.ArrayWrapper):
        assert np.all(t["mm"].data == tc["m"].data)
    elif isinstance(t["mm"], coordinates.SkyCoord):
        assert np.all(t["mm"].ra == tc["m"].ra)
        assert np.all(t["mm"].dec == tc["m"].dec)
    elif isinstance(t["mm"], coordinates.BaseRepresentationOrDifferential):
        assert np.all(representation_equal(t["mm"], tc["m"]))
    else:
        assert np.all(t["mm"] == tc["m"])


def test_represent_mixins_as_columns_unit_fix():
    """
    If the unit is invalid for a column that gets serialized this would
    cause an exception.  Fixed in #7481.
    """
    t = Table({"a": [1, 2]}, masked=True)
    t["a"].unit = "not a valid unit"
    t["a"].mask[1] = True
    serialize.represent_mixins_as_columns(t)


def test_primary_data_column_gets_description():
    """
    If the mixin defines a primary data column, that should get the
    description, format, etc., so no __info__ should be needed.
    """
    t = QTable({"a": [1, 2] * u.m})
    t["a"].info.description = "parrot"
    t["a"].info.format = "7.2f"
    tser = serialize.represent_mixins_as_columns(t)
    assert "__info__" not in tser.meta["__serialized_columns__"]["a"]
    assert tser["a"].format == "7.2f"
    assert tser["a"].description == "parrot"


def test_skycoord_with_velocity():
    # Regression test for gh-6447
    sc = SkyCoord([1], [2], unit="deg", galcen_v_sun=None)
    t = Table([sc])
    s = StringIO()
    t.write(s, format="ascii.ecsv", overwrite=True)
    s.seek(0)
    t2 = Table.read(s.read(), format="ascii.ecsv")
    assert skycoord_equal(t2["col0"], sc)


@pytest.mark.parametrize("copy", [True, False])
@pytest.mark.parametrize("table_cls", [Table, QTable])
def test_ensure_input_info_is_unchanged(table_cls, copy):
    """If a mixin input to a table has no info, it should stay that way.

    This since having 'info' slows down slicing, etc.
    See gh-11066.
    """
    q = [1, 2] * u.m
    assert "info" not in q.__dict__
    t = table_cls([q], names=["q"], copy=copy)
    assert "info" not in q.__dict__
    t = table_cls([q], copy=copy)
    assert "info" not in q.__dict__
    t = table_cls({"q": q}, copy=copy)
    assert "info" not in q.__dict__
    t["q2"] = q
    assert "info" not in q.__dict__
    sc = SkyCoord([1, 2], [2, 3], unit="deg")
    t["sc"] = sc
    assert "info" not in sc.__dict__


def test_bad_info_class():
    """Make a mixin column class that does not trigger the machinery to generate
    a pure column representation"""

    class MyArrayWrapper(ArrayWrapper):
        info = ParentDtypeInfo()

    t = Table()
    t["tm"] = MyArrayWrapper([0, 1, 2])
    out = StringIO()
    match = (
        r"failed to represent column 'tm' \(MyArrayWrapper\) as one or more Column"
        r" subclasses"
    )
    with pytest.raises(TypeError, match=match):
        represent_mixins_as_columns(t)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Extract Data Origin in VOTable.

References
----------
DataOrigin is a vocabulary described in the IVOA note: https://www.ivoa.net/documents/DataOrigin/

Notes
-----
This API retrieve Metadata from INFO in VOTable.
The information can be found at different level in a VOTable:

- global level
- resource level
- table level

Contents
--------

- Query information: Each element is considered to be unique in the VOTable.
  Information concerns publisher, date of execution, contact, request, etc.
- Dataset origin : basic provenance information.

Examples
--------
For more information, please see :ref:`DataOrigin documentation <astropy-io-votable-dataorigin>`.
"""

import astropy.io.votable.tree

__all__ = [
    "DataOrigin",
    "DatasetOrigin",
    "QueryOrigin",
    "add_data_origin_info",
    "extract_data_origin",
]


DATAORIGIN_QUERY_INFO = (
    "ivoid_service",
    "publisher",
    "server_software",
    "service_protocol",
    "request",
    "query",
    "request_date",
    "contact",
)
DATAORIGIN_INFO = (
    "ivoid",
    "citation",
    "reference_url",
    "resource_version",
    "rights_uri",
    "rights",
    "creator",
    "editor",
    "article",
    "cites",
    "is_derived_from",
    "original_date",
    "publication_date",
    "last_update_date",
)


class QueryOrigin:
    """Data class storing query execution information that generated the VOTable.

    Notes
    -----
    The Query information should be unique in the whole VOTable.
    It includes reproducibility information to execute the query again.

    Attributes
    ----------
    ivoid_service : str
        IVOID of the service that produced the VOTable (default: None)

    publisher : str
        Data centre that produced the VOTable (default: None)

    server_software : str
        Software version (default: None)

    service_protocol : str
        IVOID of the protocol through which the data was retrieved (default: None)

    request : str
        Full request URL including a query string (default: None)

    query : str
        An input query in a formal language (e.g, ADQL) (default: None)

    request_date : str
        Query execution date (default: None)

    contact : str
        Email or URL to contact publisher (default: None)

    infos : list[astropy.io.votable.tree.Info]
        list of ``<INFO>`` used by DataOrigin (default: empty list)

    """

    def __init__(self):
        self.ivoid_service = None
        self.publisher = None
        self.server_software = None
        self.service_protocol = None
        self.request = None
        self.query = None
        self.request_date = None
        self.contact = None
        self.infos = []

    def __str__(self) -> str:
        s = []
        for info_name in DATAORIGIN_QUERY_INFO:
            info = getattr(self, info_name)
            if info:
                s.append(f"{info_name}: {info}")
        return "\n".join(s)


class DatasetOrigin:
    """Data class storing the basic provenance for a Dataset.

    Notes
    -----
    DatasetOrigin is dedicated to a specific Element in a VOTable.
    These ``<INFO>`` Elements describe a Resource, a TableElement or are Global.

    Attributes
    ----------
    ivoid : list
        IVOID of underlying data collection (default: None)

    citation : list
        Dataset identifier that can be used for citation (default: None)

    reference_url : list
        Dataset landing page (default: None)

    resource_version : list
        Dataset version (default: None)

    rights_uri : list
        Licence URI (default: None)

    rights : list
        Licence or Copyright text (default: None)

    creator : list
        The person(s) mainly involved in the creation of the resource (default: None)

    editor : list
        Editor name of the reference article (default: None)

    article : list
        Bibcode or DOI of a reference article (default: None)

    cites : list
        An Identifier (ivoid, DOI, bibcode) of second resource (default: None)

    is_derived_from : list
        An Identifier (ivoid, DOI, bibcode) of second resource (default: None)

    original_date : list
        Date of the original resource from which the present resource is derived (default: None)

    publication_date : list
        Date of first publication in the data centre (default: None)

    last_update_date : list
        Last data centre update (default: None)

    infos : list[astropy.io.votable.tree.Info]
        list of ``<INFO>`` used by DataOrigin (default: None)
    """

    def __init__(self, votable_element: astropy.io.votable.tree.Element = None):
        """
        Constructor

        Parameters
        ----------
        votable_element: astropy.io.votable.tree.Element, optional
                         indicates the VOTable element
        """
        self.ivoid = None
        self.citation = None
        self.reference_url = None
        self.resource_version = None
        self.rights_uri = None
        self.rights = None
        self.creator = None
        self.editor = None
        self.article = None
        self.cites = None
        self.is_derived_from = None
        self.original_date = None
        self.publication_date = None
        self.last_update_date = None
        self.__vo_elt = votable_element
        self.infos = []

    def get_votable_element(self) -> astropy.io.votable.tree.Element:
        """
        Get the VOTable element

        Returns
        -------
        astropy.io.votable.tree.Element
        """
        return self.__vo_elt

    def __str__(self) -> str:
        s = []
        for info_name in DATAORIGIN_INFO:
            info = getattr(self, info_name)
            if info:
                s.append(f"{info_name}: {','.join(info)}")
        return "\n".join(s)


class DataOrigin:
    """Data class storing both information about query execution
       and basic provenances of datasets used to generate the VOTable.

    Attributes
    ----------
    query : QueryOrigin
        request information (default: None)

    origin : list[DatasetOrigin]
        list of DatasetOrigin (default: empty)

    Notes
    -----
    The class includes an iterator on Attribute origin.
    """

    def __init__(self):
        self.query = QueryOrigin()
        self.origin = []
        self.__it = None

    def __str__(self) -> str:
        origin_list = []
        for origin in self.origin:
            origin_list.append(str(origin))
        return str(self.query) + "\n\n" + "\n\n".join(origin_list)

    def __iter__(self):
        self.__it = -1
        return self

    def __next__(self):
        self.__it += 1
        if self.__it >= len(self.origin):
            raise StopIteration
        return self.origin[self.__it]


def __empty_dataset_origin(o: DatasetOrigin) -> bool:
    """(internal) check if DataOrigin is filled"""
    for info in DATAORIGIN_INFO:
        v = getattr(o, info)
        if v is not None:
            return False
    return True


def __extract_generic_info(
    vo_element: astropy.io.votable.tree.Element, infos: list, data_origin: DataOrigin
):
    """(internal) extract info and populate DataOrigin

    Parameters
    ----------
    vo_element : astropy.io.votable.tree.Element
        VOTable element (votable, resource or table)

    infos : list[astropy.io.votable.tree.Info]
        list of ``<INFO>``

    data_origin : DataOrigin
        DataOrigin container to fill
    """
    if not infos:
        return

    dataset_origin = DatasetOrigin(vo_element)

    for info in infos:
        info_name = info.name.lower()
        for dataorigin_info in DATAORIGIN_INFO:
            if info_name == dataorigin_info:
                dataset_origin.infos.append(info)
                att = getattr(dataset_origin, dataorigin_info)
                if att is None:
                    setattr(dataset_origin, dataorigin_info, [info.value])
                else:
                    att.append(info.value)
                break

        for query_info in DATAORIGIN_QUERY_INFO:
            if info_name == query_info:
                data_origin.query.infos.append(info)
                setattr(data_origin.query, query_info, info.value)
                break

    if not __empty_dataset_origin(dataset_origin):
        data_origin.origin.append(dataset_origin)


def __extract_dali_info(infos: list, data_origin: DataOrigin):
    """(internal) append with DALI INFO

    Parameters
    ----------
    infos : list[astropy.io.votable.tree.Info]
        iterable info

    data_origin DataOrigin
        container to fill
    """
    if not data_origin.query.service_protocol:
        for info in infos:
            info_name = info.name.lower()
            if info_name == "standardid":
                if not data_origin.query.service_protocol:
                    if data_origin.info is None:
                        data_origin.infos = []
                    data_origin.quey.infos.append(info)
                    data_origin.query.service_protocol = info.value


def __extract_info_from_table(
    table: astropy.io.votable.tree.TableElement, data_origin: DataOrigin
):
    """(internal) extract and populate dataOrigin from astropy.io.votable.tree.TableElement

    Parameters
    ----------
    table : astropy.io.votable.tree.TableElement
        Table to explore

    data_origin : DataOrigin
        container to fill.
    """
    __extract_generic_info(table, table.infos, data_origin)


def __extract_info_from_resource(
    resource: astropy.io.votable.tree.Resource,
    data_origin: DataOrigin,
    recursive: bool = True,
):
    """(internal) extract and populate dataOrigin from astropy.io.votable.tree.Resource

    Parameters
    ----------
    param resource : astropy.io.votable.tree.Resource
        Resource to explore.

    data_origin : DataOrigin
        container to fill.

    recursive : bool, optional
        make a recursive search (default: True)
    """
    __extract_generic_info(resource, resource.infos, data_origin)
    __extract_dali_info(resource.infos, data_origin)
    if recursive:
        for table in resource.tables:
            __extract_info_from_table(table, data_origin)


def __extract_info_from_votable(
    votable: astropy.io.votable.tree.VOTableFile,
    data_origin: DataOrigin,
    recursive: bool = True,
):
    """(internal) extract and populate dataOrigin from astropy.io.votable.tree.VOTableFile

    Parameters
    ----------
    votable : astropy.io.votable.tree.VOTableFile
        VOTableFile to explore

    data_origin : DataOrigin
        container to fill

    recursive : bool, optional
        make a recursive search (default: True)
    """
    __extract_generic_info(votable, votable.infos, data_origin)
    if recursive:
        for resource in votable.resources:
            __extract_info_from_resource(resource, data_origin)


def extract_data_origin(vot_element: astropy.io.votable.tree.Element) -> DataOrigin:
    """Extract DataOrigin in a VO element

    Parameters
    ----------
    vot_element : astropy.io.votable.tree.Info
        VOTable Element to explore

    Returns
    -------
    DataOrigin

    Raises
    ------
    TypeError
        input ``vot_element`` type is not supported
    """
    data_origin = DataOrigin()
    if isinstance(vot_element, astropy.io.votable.tree.VOTableFile):
        __extract_info_from_votable(vot_element, data_origin)
    elif isinstance(vot_element, astropy.io.votable.tree.Resource):
        __extract_info_from_resource(vot_element, data_origin)
    elif isinstance(vot_element, astropy.io.votable.tree.TableElement):
        __extract_info_from_table(vot_element, data_origin)
    else:
        raise TypeError("input vot_element type is not supported.")

    return data_origin


def add_data_origin_info(
    vot_element: astropy.io.votable.tree.Element,
    info_name: str,
    info_value: str,
    content: str | None = None,
) -> None:
    """Update VOTable element with information compatible
       with DataOrigin vocabulary.

    Notes
    -----
    The function checks information name and adds the
    VOTable element with a new ``<INFO>``.

    Parameters
    ----------
    vot_element : astropy.io.votable.tree.Element
        VOTable element where to add the information

    info_name : str
        Attribute name (see DATAORIGIN_INFO, DATAORIGIN_QUERY_INFO)

    info_value : str
        value

    content : str, optional
        Content in ``<INFO>`` (default: None)

    Raises
    ------
    TypeError
        input type not managed or information name not recognized
    ValueError
        ``info_name`` already exists in ``vot_element``
    ValueError
        ``info_name`` is an unknown DataOrigin name.
    """
    if info_name in DATAORIGIN_INFO:
        if not isinstance(
            vot_element,
            (
                astropy.io.votable.tree.VOTableFile,
                astropy.io.votable.tree.Resource,
                astropy.io.votable.tree.TableElement,
            ),
        ):
            raise TypeError("Unsupported vot_element type.")

    elif info_name in DATAORIGIN_QUERY_INFO:
        if not isinstance(vot_element, astropy.io.votable.tree.VOTableFile):
            raise TypeError(
                "Bad type of vot_element: this information needs VOTableFile."
            )

        for info in vot_element.get_infos_by_name(info_name):
            raise ValueError(f"QueryOrigin {info_name} already exists")

    else:
        raise ValueError("Unknown DataOrigin info name.")

    new_info = astropy.io.votable.tree.Info(name=info_name, value=info_value)
    if content:
        new_info.content = content
    vot_element.infos.extend([new_info])



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Functions related to Python runtime introspection."""

import inspect
import os
import sys
from importlib import import_module, metadata
from importlib.metadata import packages_distributions
from types import FrameType, ModuleType
from typing import Literal

from packaging.version import Version

from .decorators import deprecated

__all__ = ["find_current_module", "isinstancemethod", "minversion", "resolve_name"]

__doctest_skip__ = ["find_current_module"]


@deprecated(
    since="7.0", alternative="importlib (e.g. importlib.import_module for modules)"
)
def resolve_name(name: str, *additional_parts: str) -> object:
    """Resolve a name like ``module.object`` to an object and return it.

    This ends up working like ``from module import object`` but is easier
    to deal with than the `__import__` builtin and supports digging into
    submodules.

    Parameters
    ----------
    name : `str`
        A dotted path to a Python object--that is, the name of a function,
        class, or other object in a module with the full path to that module,
        including parent modules, separated by dots.  Also known as the fully
        qualified name of the object.

    additional_parts : iterable, optional
        If more than one positional arguments are given, those arguments are
        automatically dotted together with ``name``.

    Raises
    ------
    `ImportError`
        If the module or named object is not found.
    """
    additional_parts = ".".join(additional_parts)

    if additional_parts:
        name = name + "." + additional_parts

    parts = name.split(".")

    if len(parts) == 1:
        # No dots in the name--just a straight up module import
        cursor = 1
        fromlist = []
    else:
        cursor = len(parts) - 1
        fromlist = [parts[-1]]

    module_name = parts[:cursor]

    while cursor > 0:
        try:
            ret = __import__(".".join(module_name), fromlist=fromlist)
            break
        except ImportError:
            if cursor == 0:
                raise
            cursor -= 1
            module_name = parts[:cursor]
            fromlist = [parts[cursor]]
            ret = ""

    for part in parts[cursor:]:
        try:
            ret = getattr(ret, part)
        except AttributeError:
            raise ImportError(name)

    return ret


def minversion(module: ModuleType | str, version: str, inclusive: bool = True) -> bool:
    """
    Returns `True` if the specified Python module satisfies a minimum version
    requirement, and `False` if not.

    Parameters
    ----------
    module : module or `str`
        An imported module of which to check the version, or the name of
        that module (in which case an import of that module is attempted--
        if this fails `False` is returned).

    version : `str`
        The version as a string that this module must have at a minimum (e.g.
        ``'0.12'``).

    inclusive : `bool`
        The specified version meets the requirement inclusively (i.e. ``>=``)
        as opposed to strictly greater than (default: `True`).

    Examples
    --------
    >>> import astropy
    >>> minversion(astropy, '0.4.4')
    True
    """
    if inspect.ismodule(module):
        module_name = module.__name__
        module_version = getattr(module, "__version__", None)
    elif isinstance(module, str):
        module_name = module
        module_version = None
        try:
            module = import_module(module_name)
        except ImportError:
            return False
    else:
        raise ValueError(
            "module argument must be an actual imported "
            "module, or the import name of the module; "
            f"got {repr(module)}"
        )

    if module_version is None:
        try:
            module_version = metadata.version(module_name)
        except metadata.PackageNotFoundError:
            # Maybe the distribution name is different from package name.
            # Calling packages_distributions is costly so we do it only
            # if necessary, as only a few packages don't have the same
            # distribution name.
            dist_names = packages_distributions()
            module_version = metadata.version(dist_names[module_name][0])

    if inclusive:
        return Version(module_version) >= Version(version)
    else:
        return Version(module_version) > Version(version)


def find_current_module(
    depth: int = 1, finddiff: bool | list[Literal[True] | str | ModuleType] = False
) -> ModuleType | None:
    """
    Determines the module/package from which this function is called.

    This function has two modes, determined by the ``finddiff`` option. it
    will either simply go the requested number of frames up the call
    stack (if ``finddiff`` is False), or it will go up the call stack until
    it reaches a module that is *not* in a specified set.

    Parameters
    ----------
    depth : int
        Specifies how far back to go in the call stack (0-indexed, so that
        passing in 0 gives back `astropy.utils.misc`).
    finddiff : bool or list
        If False, the returned ``mod`` will just be ``depth`` frames up from
        the current frame. Otherwise, the function will start at a frame
        ``depth`` up from current, and continue up the call stack to the
        first module that is *different* from those in the provided list.
        In this case, ``finddiff`` can be a list of modules or modules
        names. Alternatively, it can be True, which will use the module
        ``depth`` call stack frames up as the module the returned module
        most be different from.

    Returns
    -------
    mod : module or None
        The module object or None if the package cannot be found. The name of
        the module is available as the ``__name__`` attribute of the returned
        object (if it isn't None).

    Raises
    ------
    ValueError
        If ``finddiff`` is a list with an invalid entry.

    Examples
    --------
    The examples below assume that there are two modules in a package named
    ``pkg``. ``mod1.py``::

        def find1():
            from astropy.utils import find_current_module
            print find_current_module(1).__name__
        def find2():
            from astropy.utils import find_current_module
            cmod = find_current_module(2)
            if cmod is None:
                print 'None'
            else:
                print cmod.__name__
        def find_diff():
            from astropy.utils import find_current_module
            print find_current_module(0,True).__name__

    ``mod2.py``::

        def find():
            from .mod1 import find2
            find2()

    With these modules in place, the following occurs::

        >>> from pkg import mod1, mod2
        >>> from astropy.utils import find_current_module
        >>> mod1.find1()
        pkg.mod1
        >>> mod1.find2()
        None
        >>> mod2.find()
        pkg.mod2
        >>> find_current_module(0)
        <module 'astropy.utils.misc' from 'astropy/utils/misc.py'>
        >>> mod1.find_diff()
        pkg.mod1

    """
    frm = inspect.currentframe()
    for _ in range(depth):
        frm = frm.f_back
        if frm is None:
            return None

    if finddiff:
        currmod = _get_module_from_frame(frm)
        if finddiff is True:
            diffmods = [currmod]
        else:
            diffmods = []
            for fd in finddiff:
                if inspect.ismodule(fd):
                    diffmods.append(fd)
                elif isinstance(fd, str):
                    diffmods.append(import_module(fd))
                elif fd is True:
                    diffmods.append(currmod)
                else:
                    raise ValueError("invalid entry in finddiff")

        while frm:
            frmb = frm.f_back
            modb = _get_module_from_frame(frmb)
            if modb not in diffmods:
                return modb
            frm = frmb
    else:
        return _get_module_from_frame(frm)


def _get_module_from_frame(frm: FrameType) -> ModuleType | None:
    """Uses inspect.getmodule() to get the module that the current frame's
    code is running in.

    However, this does not work reliably for code imported from a zip file,
    so this provides a fallback mechanism for that case which is less
    reliable in general, but more reliable than inspect.getmodule() for this
    particular case.
    """
    mod = inspect.getmodule(frm)
    if mod is not None:
        return mod

    # Check to see if we're importing from a bundle file. First ensure that
    # __file__ is available in globals; this is cheap to check to bail out
    # immediately if this fails

    if "__file__" in frm.f_globals and "__name__" in frm.f_globals:
        filename = frm.f_globals["__file__"]

        # Using __file__ from the frame's globals and getting it into the form
        # of an absolute path name with .py at the end works pretty well for
        # looking up the module using the same means as inspect.getmodule

        if filename[-4:].lower() in (".pyc", ".pyo"):
            filename = filename[:-4] + ".py"
        filename = os.path.realpath(os.path.abspath(filename))
        if filename in inspect.modulesbyfile:
            return sys.modules.get(inspect.modulesbyfile[filename])

        # On Windows, inspect.modulesbyfile appears to have filenames stored
        # in lowercase, so we check for this case too.
        if filename.lower() in inspect.modulesbyfile:
            return sys.modules.get(inspect.modulesbyfile[filename.lower()])

    # Otherwise there are still some even trickier things that might be possible
    # to track down the module, but we'll leave those out unless we find a case
    # where it's really necessary.  So return None if the module is not found.
    return None


@deprecated(since="6.1")
def find_mod_objs(modname, onlylocals=False):
    """Returns all the public attributes of a module referenced by name.

    .. note::
        The returned list *not* include subpackages or modules of
        ``modname``, nor does it include private attributes (those that
        begin with '_' or are not in `__all__`).

    Parameters
    ----------
    modname : str
        The name of the module to search.
    onlylocals : bool or list of str
        If `True`, only attributes that are either members of ``modname`` OR
        one of its modules or subpackages will be included. If it is a list
        of strings, those specify the possible packages that will be
        considered "local".

    Returns
    -------
    localnames : list of str
        A list of the names of the attributes as they are named in the
        module ``modname`` .
    fqnames : list of str
        A list of the full qualified names of the attributes (e.g.,
        ``astropy.utils.introspection.find_mod_objs``). For attributes that are
        simple variables, this is based on the local name, but for functions or
        classes it can be different if they are actually defined elsewhere and
        just referenced in ``modname``.
    objs : list of objects
        A list of the actual attributes themselves (in the same order as
        the other arguments)

    """
    mod = import_module(modname)

    if hasattr(mod, "__all__"):
        pkgitems = [(k, getattr(mod, k)) for k in mod.__all__]
    else:
        pkgitems = [(k, getattr(mod, k)) for k in dir(mod) if k[0] != "_"]

    # filter out modules and pull the names and objs out
    ismodule = inspect.ismodule
    localnames = [k for k, v in pkgitems if not ismodule(v)]
    objs = [v for k, v in pkgitems if not ismodule(v)]

    # fully qualified names can be determined from the object's module
    fqnames = []
    for obj, lnm in zip(objs, localnames):
        if hasattr(obj, "__module__") and hasattr(obj, "__name__"):
            fqnames.append(obj.__module__ + "." + obj.__name__)
        else:
            fqnames.append(modname + "." + lnm)

    if onlylocals:
        if onlylocals is True:
            onlylocals = [modname]
        valids = [any(fqn.startswith(nm) for nm in onlylocals) for fqn in fqnames]
        localnames = [e for i, e in enumerate(localnames) if valids[i]]
        fqnames = [e for i, e in enumerate(fqnames) if valids[i]]
        objs = [e for i, e in enumerate(objs) if valids[i]]

    return localnames, fqnames, objs


# Note: I would have preferred call this is_instancemethod, but this naming is
# for consistency with other functions in the `inspect` module
@deprecated(since="6.1")
def isinstancemethod(cls, obj):
    """
    Returns `True` if the given object is an instance method of the class
    it is defined on (as opposed to a `staticmethod` or a `classmethod`).

    This requires both the class the object is a member of as well as the
    object itself in order to make this determination.

    Parameters
    ----------
    cls : `type`
        The class on which this method was defined.
    obj : `object`
        A member of the provided class (the membership is not checked directly,
        but this function will always return `False` if the given object is not
        a member of the given class).
    """
    if not inspect.isfunction(obj):
        return False

    # Unfortunately it seems the easiest way to get to the original
    # staticmethod object is to look in the class's __dict__, though we
    # also need to look up the MRO in case the method is not in the given
    # class's dict
    name = obj.__name__
    for basecls in cls.mro():  # This includes cls
        if name in basecls.__dict__:
            return not isinstance(basecls.__dict__[name], staticmethod)

    # This shouldn't happen, though this is the most sensible response if
    # it does.
    raise AttributeError(name)
